"""
Functions and classes for handling time

Hierarchy of TimePars:
TimePar  # All time parameters
├── Dur  # All durations, units of *time*
│   ├── days  # Duration with units of days
│   ├── weeks
│   ├── months
│   ├── years
│   └── DateDur  # Calendar durations
└── Rate  # All rates, units of *per time*
    ├── rateperday  # Number of events happening per day
    ├── rateperweek
    ├── ratepermonth
    ├── rateperyear
    ├── TimeProb  # Probability of an event happening in a given time
    │   ├── perday # Probability of an event happening in a day
    │   ├── perweek
    │   ├── permonth
    │   └── peryear
    └── RateProb  # Instantaneous probability of an event happening
"""
import sciris as sc
import numpy as np
import pandas as pd
import starsim as ss

# General classes; specific classes are listed below
__all__ = ['date', 'TimePar', 'Dur', 'DateDur', 'Rate', 'TimeProb', 'RateProb']

#%% Base classes
class DateArray(np.ndarray):
    """ Lightweight wrapper for an array of dates """
    def __new__(cls, arr=None):
        if arr is None:
            arr = np.array([])
        if isinstance(arr, np.ndarray): # Shortcut to typical use case, where the input is an array
            return arr.view(cls)
        else:
            errormsg = f'Argument must be an array, not {type(arr)}'
            raise TypeError(errormsg)


class date(pd.Timestamp):
    """
    Define a point in time, based on `pd.Timestamp`

    Args:
        date (int/float/str/datetime): Any type of date input (ints and floats will be interpreted as years)
        kwargs (dict): passed to pd.Timestamp()

    **Examples**:

        ss.date(2020) # Returns <2020-01-01>
        ss.date(year=2020) # Returns <2020-01-01>
        ss.date(year=2024.75) # Returns <2024-10-01>
        ss.date('2024-04-04') # Returns <2024-04-04>
        ss.date(year=2024, month=4, day=4) # Returns <2024-04-04>
    """

    def __new__(cls, *args, **kwargs):
        """ Check if a year was supplied, and preprocess it; complex due to pd.Timestamp implementation """
        single_year_arg = False
        if len(args) == 1:
            arg = args[0]
            if arg is None:
                return pd.Timestamp(None)
            elif isinstance(arg, pd.Timestamp):
                return cls._reset_class(sc.dcp(arg))
            elif sc.isnumber(arg): # e.g. 2020
                single_year_arg = True
            elif isinstance(arg, ss.DateDur): # e.g. ss.DateDur(years=2020)
                kwargs.update(arg.to_dict())
            elif isinstance(arg, ss.Dur): # e.g. ss.years(2020)
                arg = arg.years
                single_year_arg = True

        year_kwarg = len(args) == 0 and len(kwargs) == 1 and 'year' in kwargs
        if single_year_arg:
            return cls.from_year(arg)
        if year_kwarg:
            return cls.from_year(kwargs['year'])

        # Otherwise, proceed as normal
        out = super(date, cls).__new__(cls, *args, **kwargs)
        out = cls._reset_class(out)
        return out

    @classmethod
    def _reset_class(cls, obj):
        """ Manually reset the class from pd.Timestamp to ss.date """
        obj.__class__ = date
        return obj

    def __reduce__(self):
        # This function is internally used when pickling (rather than getstate/setstate)
        # due to Pandas implementing C functions for this. We can wrap the Pandas unpickler
        # with a function that wraps calls _reset_class to support unpickling date
        # objects at the instance level
        unpickling_func, args = super().__reduce__()
        return (self.__class__._rebuild, (self.__class__, unpickling_func, args))

    @staticmethod
    def _rebuild(cls, unpickling_func, args):
        out = unpickling_func(*args)
        out = cls._reset_class(out)
        return out

    def __repr__(self, bracket=True):
        """ Show the date in brackets, e.g. <2024.04.04> """
        _ = ss.options.date_sep
        y = f'{self.year:04d}'
        m = f'{self.month:02d}'
        d = f'{self.day:02d}'
        string = y + _ + m + _ + d

        if self._time_repr != '00:00:00':
            string += ' ' + self._time_repr[:8]

        if bracket:
            string = '<' + string + '>'
        return string

    def __str__(self):
        """ Like repr, but just the date, e.g. 2024.04.04 """
        return self.__repr__(bracket=False)

    def replace(self, *args, **kwargs):
        """ Returns a new ss.date(); pd.Timestamp is immutable """
        out = super().replace(*args, **kwargs)
        out = self.__class__._reset_class(out)
        return out

    def disp(self, **kwargs):
        """ Show the full object """
        return sc.pr(self, **kwargs)

    @classmethod
    def from_year(cls, year):
        """
        Convert an int or float year to a date.

        **Examples**:

            ss.date.from_year(2020) # Returns <2020-01-01>
            ss.date.from_year(2024.75) # Returns <2024-10-01>
        """
        if isinstance(year, int):
            return cls(year=year, month=1, day=1)
        else:
            year_start = pd.Timestamp(year=int(year), month=1, day=1)
            year_end = pd.Timestamp(year=int(year) + 1, month=1, day=1)
            return cls._reset_class(year_start + year % 1 * (year_end - year_start))

    def to_year(self):
        """
        Convert a date to a floating-point year

        **Examples**:

            ss.date('2020-01-01').to_year() # Returns 2020.0
            ss.date('2024-10-01').to_year() # Returns 2024.7486
        """
        year_start = pd.Timestamp(year=self.year,month=1,day=1).timestamp()
        year_end = pd.Timestamp(year=self.year+1,month=1,day=1).timestamp()
        return self.year + (self.timestamp()-year_start)/(year_end-year_start)

    def __float__(self):
        return self.to_year()

    @property
    def years(self):
        # This matches Dur.years
        return self.to_year()

    def to_pandas(self):
        """ Convert to a standard pd.Timestamp instance """
        return pd.Timestamp(self.to_numpy()) # Need to convert to NumPy first or it doesn't do anything

    def _timestamp_add(self, other):
        """ Uses pd.Timestamp's __add__ and avoids creating duplicate objects """
        orig_class = self.__class__
        self.__class__ = pd.Timestamp
        out = orig_class._reset_class(self + other)
        orig_class._reset_class(self)
        return out

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            return np.vectorize(self.__add__)(other)
        elif isinstance(other, DateDur):
            return self._timestamp_add(other.value)
        elif isinstance(other, years):
            return date(self.to_year() + other.value)
        elif isinstance(other, pd.DateOffset):
            return self._timestamp_add(other)
        elif isinstance(other, pd.Timestamp):
            raise TypeError('Cannot add a date to another date')
        elif sc.isnumber(other):
            raise TypeError(f'Attempted to add a number ({other}) to a date, which is not supported. Only durations can be added to dates e.g., "ss.years({other})" or "ss.days({other})"')
        else:
            raise TypeError(f'Attempted to add an instance of {other.__class__.__name__} to a date, which is not supported. Only durations can be added to dates.')

    def __sub__(self, other):

        if isinstance(other, np.ndarray):
            return np.vectorize(self.__sub__)(other)

        if isinstance(other, DateDur):
            return date(self.to_pandas() - other.value)
        elif isinstance(other, years):
            return date(self.to_year() - other.value)
        elif isinstance(other, pd.DateOffset):
            return date(self.to_pandas() - other)
        elif isinstance(other, date):
            return years(self.years-other.years)
        else:
            errormsg = f'Unsupported type {type(other)}'
            raise TypeError(errormsg)

    def __radd__(self, other): return self.__add__(other)
    def __rsub__(self, other): return self.__sub__(other) # TODO: check if this should be reversed

    def __lt__(self, other):
        if sc.isnumber(other):
            return self.to_year() < other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__lt__)(other)
        else:
            return super().__lt__(other)

    def __gt__(self, other):
        if sc.isnumber(other):
            return self.to_year() > other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__gt__)(other)
        else:
            return super().__gt__(other)

    def __le__(self, other):
        if sc.isnumber(other):
            return self.to_year() <= other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__le__)(other)
        else:
            return super().__le__(other)

    def __ge__(self, other):
        if sc.isnumber(other):
            return self.to_year() >= other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__ge__)(other)
        else:
            return super().__ge__(other)

    def __eq__(self, other):
        if sc.isnumber(other):
            return self.to_year() == other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__eq__)(other)
        else:
            return super().__eq__(other)

    def __ne__(self, other):
        if sc.isnumber(other):
            return self.to_year() != other
        elif isinstance(other, np.ndarray):
            return np.vectorize(self.__ne__)(other)
        else:
            return super().__ne__(other)

    def __hash__(self):
        # As equality with float years is implemented, this allows interoperability with dicts that use float year keys
        return hash(self.to_year())

    # Convenience methods based on application usage

    @classmethod
    def from_array(cls, array):
        """
        Convert an array of float years into an array of date instances

        Args:
            array: An array of float years

        Returns:
            An array of date instances
        """
        return DateArray(np.vectorize(cls)(array))

    @classmethod
    def arange(cls, low, high, step=1):
        """
        Construct an array of dates

        Functions similarly to np.arange, but returns date objects

        Example usage:

        >>> date.arange(2020, 2025)
            array([<2020.01.01>, <2021.01.01>, <2022.01.01>, <2023.01.01>,
                   <2024.01.01>], dtype=object)

        Args:
            low: Lower bound - can be a date or a numerical year
            high: Upper bound - can be a date or a numerical year
            step: Assumes 1 calendar year steps by default

        Returns:
            An array of date instances
        """
        # Convert this first
        if isinstance(step, ss.Dur):
            step = DateDur(step)

        if isinstance(step, ss.DateDur):
            if not isinstance(low, date):
                low = cls(low)
            if not isinstance(high, date):
                high = cls(high)

            tvec = []
            t = low
            while t <= high:
                tvec.append(t)
                t += step
            return DateArray(np.array(tvec))
        elif sc.isnumber(step):
            low = low.years if isinstance(low, date) else low
            high = high.years if isinstance(high, date) else high
            return cls.from_array(np.arange(low, high, step))
        else:
            errormsg = f'Cannot construct date range from {low = }, {high = }, and {step = }. Expecting ss.date(), ss.Dur(), or numbers as inputs.'
            raise TypeError(errormsg)

    def to_json(self):
        """ Returns a JSON representation of the date """
        out = {'ss.date': str(self)}
        return out

    @staticmethod
    def from_json(json):
        """ Reconstruct a date from a JSON; reverse of `to_json()` """
        if not (isinstance(json, dict) and len(json) == 1 and list(json.keys())[0] == 'ss.date'):
            errormsg = f'Expecting a dict with a single key "ss.date", not {json}'
            raise ValueError(errormsg)
        return date(json['ss.date'])



class UnitFactors(sc.dictobj):
    def __init__(self, base, unit):
        unit_items = base.copy()
        for k,v in unit_items.items():
            unit_items[k] /= base[f'{unit}s'] # e.g. 'year' -> 'years'
        self.unit = unit
        self.items = unit_items
        self.values = np.array(list(unit_items.values())) # Precompute for computational efficiency
        self.keys = list(unit_items.keys())
        return

    def __repr__(self):
        return f'{self.unit}:\n{repr(sc.objdict(self.items))}'

    def disp(self):
        return sc.pr(self)


class Factors(sc.dictobj):
    """ Define time unit conversion factors """
    unit_keys = ['year', 'month', 'week', 'day']

    def __init__(self):
        base = sc.dictobj( # Use dictobj since slightly faster than objdict
            years   = 1, # Note 'years' instead of 'year', since referring to a quantity instead of a unit
            months  = 12,
            weeks   = 365/7, # If 52, then day/week conversion is incorrect
            days    = 365,
            hours   = 365*24,
            minutes = 365*24*60,
            seconds = 365*24*60*60,
        )
        for key in self.unit_keys:
            unit_factors = UnitFactors(base, key)
            self[key] = unit_factors
        return

    def __repr__(self):
        string = ''
        for key in self.unit_keys:
            string += 'factors.'
            entry = repr(self[key])
            entry = sc.indent(text=entry, n=4).lstrip() # Don't indent the first line
            string += entry
        return string

# Preallocate for performance
factors = Factors()

class TimePar:
    """ Parent class for all TimePars -- Dur, Rate, etc. """
    base = None # e.g. 'year', 'day', etc
    timepar_type = None # 'dur' or 'rate'
    timepar_subtype = None # e.g. 'datedur' or 'timeprob'

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._set_factors()
        return

    @classmethod
    def _set_factors(cls):
        if cls.base is not None:
            cls.basekey = cls.base + 's' # 'year' -> 'years'
            cls.factors = factors[cls.base].items
            cls.factor_keys = factors[cls.base].keys
            cls.factor_vals = factors[cls.base].values
        return

    def __setattrr__(self, attr, value):
        if object.__getattribute__(self, '_locked'):
            errormsg = f'Cannot set attributes of {self}; object is read-only'
            raise AttributeError(errormsg)
        else:
            super().__setattrr__(self, attr, value)
            return

    def __getitem__(self, index):
        """ For indexing and slicing, e.g. TimePar[inds] """
        if isinstance(self.value, np.ndarray):
            return self.value[index]
        else:
            errormsg = f'Can only index {type(self)} if value is an array, not a scalar'
            raise TypeError(errormsg)

    def __setitem__(self, index, value):
        """ For indexing and slicing, e.g. TimePar[inds] """
        if isinstance(self.value, np.ndarray):
            self.value[index] = value
            return
        else:
            errormsg = f'Can only index {type(self)} if value is an array, not a scalar'
            raise TypeError(errormsg)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """ Disallow array operations by default, as they create arrays of objects (basically lists) """
        a,b = inputs
        a_b = True
        if b is self: # Swap order if needed
            b,a = a,b
            a_b = True

        if   ufunc == np.add:      return self.__add__(b)
        elif ufunc == np.subtract: return self.__sub__(b) if a_b else self.__rsub__(b)
        elif ufunc == np.multiply: return self.__mul__(b)
        elif ufunc == np.divide:   return self.__truediv__(b) if a_b else self.__rtruediv__(b)
        else:
            self.array_mul_error(ufunc)
        return

    def array_mul_error(self, ufunc=None):
        ufuncstr = f"{ufunc} " if ufunc is not None else ''
        errormsg = f'Cannot perform array operation {ufuncstr}on {self}: this would create an array of {self.__class__} objects, which is very inefficient. '
        errormsg += 'Are you missing parentheses? If you convert to a unitless quantity by multiplying a rate and a duration, then you can multiply by an array. '
        errormsg += 'If you really want to do this, use an explicit loop instead.'
        raise TypeError(errormsg)

    def to(self, unit):
        """ Convert this TimePar to one of a different class """
        return class_map[self.timepar_type][unit](self)

    @classmethod
    def to_base(cls, other):
        """ Convert another TimePar object to this TimePar's base units """
        try:
            return getattr(other, cls.basekey) # e.g. other.years
        except AttributeError as e:
            errormsg = f'Cannot get property {cls.basekey} from object {other}. Is it a TimePar?'
            raise AttributeError(errormsg) from e

    def mutate(self, unit):
        """ Mutate a TimePar in place to a new base unit -- see `TimePar.to()` to return a new instance (much more common) """
        self.base = unit
        self._set_factors()
        self.__class__ = class_map[self.timepar_type][unit] # Mutate the class in place
        return self


class Dur(TimePar):
    """
    Base class for durations

    Note: this class should not be used by the user directly; instead, use ss.years(),
    ss.days(), etc.

    Note that although they are different classes, `ss.Dur` objects can be modified
    in place if needed via the `ss.Dur.mutate()` method.
    """
    base = None
    timepar_type = 'dur'
    timepar_subtype = 'dur'

    def __new__(cls, *args, **kwargs):
        # Return
        if cls is Dur:
            if args:
                if isinstance(args[0], (pd.DateOffset, DateDur)):
                    return super().__new__(DateDur)
                elif isinstance(args[0], years):
                    return super().__new__(years)
                else:
                    assert len(args) == 1, f'Dur must be instantiated with only 1 arg (which is in years), or keyword arguments. {len(args)} args were given.'
                    return super().__new__(years)
            else:
                return super().__new__(DateDur) # TODO: do not make the default, but needs new classes
        return super().__new__(cls)

    def __init__(self, value=1):
        """
        Construct a value-based duration

        Args:
            value (float/`ss.Dur`): the value to use
        """
        super().__init__()
        if isinstance(value, Dur):
            self.value = self.to_base(value)
        else:
            self.value = value

    def __repr__(self):
        if self.value == 1:
            return f'{self.base}()'
        else:
            return f'{self.basekey}({self.value})'

    # NB. Durations are considered to be equal if their year-equivalent duration is equal
    # That would mean that Dur(years=1)==Dur(1) returns True - probably less confusing than having it return False?
    def __hash__(self):
        return hash(self.years)

    def __float__(self):
        return float(self.value)

    @property
    def years(self):
        return self.value*self.factors.years # Needs to return float so matplotlib can plot it correctly

    @property
    def months(self):
        return self.value*self.factors.months # Needs to return float so matplotlib can plot it correctly

    @property
    def weeks(self):
        return self.value*self.factors.weeks # Needs to return float so matplotlib can plot it correctly

    @property
    def days(self):
        return self.value*self.factors.day # Needs to return float so matplotlib can plot it correctly

    def to_numpy(self):
        return sc.toarray(self.value)

    def __add__(self, other):
        if isinstance(other, Dur):
            return self.__class__(self.value + self.to_base(other))
        elif isinstance(other, date): # If adding to a date, convert to years
            return date.from_year(other.to_year() + self.years)
        elif isinstance(other, DateArray):
            return DateArray(np.vectorize(self.__add__)(other))
        else:
            return self.__class__(self.value + other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Dur):
            out = self.__class__(self.value - self.to_base(other))
        elif isinstance(other, date):
            return date.from_year(other.to_year() - self.years)
        else:
            out = self.__class__(self.value - other)
            if sc.isnumber(out) and out < 0:
                warnmsg = f'Subtracting {self} and {other} yields {out}. Durations are rarely negative; are you sure this is intentional?'
                ss.warn(warnmsg)
        return out

    def __mul__(self, other):
        if isinstance(other, Rate):
            return NotImplemented # Delegate to Rate.__rmul__
        elif isinstance(other, Dur):
            raise Exception('Cannot multiply a duration by a duration')
        elif isinstance(other, date):
            raise Exception('Cannot multiply a duration by a date')
        return self.__class__(self.value*other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Dur):
            return self.years/other.years
        elif isinstance(other, Rate):
            raise Exception('Cannot divide a duration by a rate')
        else:
            return self.__class__(self.value / other)

    def __neg__(self):
        return -1*self

    def __lt__(self, other):
        try:
            return self.years < other.years
        except:
            return self.years < other

    def __gt__(self, other):
        try:
            return self.years > other.years
        except:
            return self.years > other

    def __le__(self, other):
        try:
            return self.years <= other.years
        except:
            return self.years <= other

    def __ge__(self, other):
        try:
            return self.years >= other.years
        except:
            return self.years >= other

    def __eq__(self, other):
        try:
            return self.years == other.years
        except:
            return self.years == other

    def __ne__(self, other):
        try:
            return self.years != other.years
        except:
            return self.years != other

    def __rtruediv__(self, other):
        # If a Dur is divided by a Dur then we will call __truediv__
        # If a float is divided by a Dur, then we should return a rate
        # If a rate is divided by a Dur, then we will call Rate.__truediv__
        # We also need divide the duration by the numerator when calculating the rate
        if self.years == 0:
            raise ZeroDivisionError('Cannot divide by a duration of zero') # TODO: consider Rate(Dur(np.inf))
        elif (sc.isnumber(other) and other == 0) or isinstance(other, Dur) and other.years == 0:
            return Rate(0)
        else:
            return Rate(other, self)

    def __abs__(self):
        return self.__class__(abs(self.value))

    @classmethod
    def arange(cls, low, high, step): # TODO: remove? ss.Dur(arr) is preferable to array(ss.Dur) for performance
        """
        Construct an array of Dur instances

        For this function, the low, high, and step must ALL be specified, and they must
        all be Dur instances. Mixing Dur types (years and DateDur) is permitted.

        Args:
            low (ss.Dur): Starting point, e.g., ss.years(0)
            high (ss.Dur): Ending point, e.g. ss.years(20)
            step (ss.Dur): Step size, e.g. ss.years(2)
        """
        args = [low, high, step]
        if not all([isinstance(arg, ss.Dur) for arg in args]):
            errormsg = f'All inputs must be ss.Dur, not {args}'
            raise TypeError(errormsg)

        tvec = []
        t = low
        while t <= high:
            tvec.append(t)
            t += step
        return np.array(tvec)


class DateDur(Dur):
    """ Date based duration e.g., if requiring a week to be 7 calendar days later """
    base = 'year' # DateDur uses 'year' as the default unit for conversion
    timepar_type = 'dur'
    timepar_subtype = 'datedur'

    def __init__(self, *args, **kwargs):
        """
        Create a date-based duration

        Supported arguments are
            - Single positional argument
                - A float number of years
                - A pd.DateOffset
                - A Dur instance
            - Keyword arguments
                - Argument names that match time periods in `ss.time.factors` e.g., 'years', 'months'
                  supporting floating point values. If the value is not an integer, the values
                  will be cascaded to smaller units using `ss.time.factors`

        If no arguments are specified, a DateDur with zero duration will be produced

        Args:
            args:
            kwargs:
        """
        super().__init__()
        if args:
            assert not kwargs, 'DateDur must be instantiated with only 1 arg (which is in years), or keyword arguments.'
            assert len(args) == 1, f'DateDur must be instantiated with only 1 arg (which is in years), or keyword arguments. {len(args)} args were given.'
            if isinstance(args[0], pd.DateOffset):
                self.value = self._round_duration(args[0])
            elif isinstance(args[0], DateDur):
                self.value = sc.dcp(args[0].value) # pd.DateOffset is immutable so this should be OK
            elif isinstance(args[0], Dur):
                self.value = self._round_duration({'years': args[0].years})
            elif sc.isnumber(args[0]):
                self.value = self._round_duration({'years': args[0]})
            else:
                errormsg = f'Unsupported input {args}.\nExpecting number, ss.Dur, ss.DateDur, or pd.DateOffset'
                raise TypeError(errormsg)
        else:
            self.value = self._round_duration(kwargs)

    def __float__(self):
        return float(self.years)

    @classmethod
    def _as_array(cls, dateoffset):
        """
        Return array representation of a pd.DateOffset

        In the array representation, each element corresponds to one of the time units
        in `ss.time.factors` e.g., a 1 year, 2 week duration would be [1, 0, 2, 0, 0, 0, 0, 0]

        Args:
            dateoffset: A pd.DateOffset object

        Returns:
            A numpy array with as many elements as `ss.time.factors`
        """
        return np.array([dateoffset.kwds.get(k, 0) for k in cls.factor_keys])

    @classmethod
    def _as_args(cls, x):
        """
        Return a dictionary representation of a DateOffset or DateOffset array

        This function takes in either a pd.DateOffset, or the result of `_as_array`, and returns
        a dictionary with the same keys as `ss.time.factors` and the values from the input. The
        output of this function can be passed to the `Dur()` constructor to create a new DateDur object.

        Args:
            x: A pd.DateOffset or an array with the same number of elements as `ss.time.factors`

        Returns:
            Dictionary with keys from `ss.time.factors` and values from the input
        """
        if isinstance(x, pd.DateOffset):
            x = cls._as_array(x)
        return {k: v for k, v in zip(cls.factor_keys, x)}

    def to_array(self):
        """ Convert to a Numpy array (NB, different than to_numpy() which converts to fractional years """
        return self._as_array(self.value)

    def to_numpy(self):
        # TODO: This is a quick fix to get plotting to work, but this should do something more sensible
        return self.years

    def to_dict(self):
        """ Convert to a dictionary """
        return self._as_args(self.to_array())

    @property
    def years(self):
        """
        Return approximate conversion into years

        This allows interoperability with years objects (which would typically be expected to
        occur if module parameters have been entered with `DateDur` durations, but the simulation
        timestep is in `years` units).

        The conversion is based on `ss.time.factors` which defines the conversion from each time unit
        to the next

        Returns:
            A float representing the duration in years
        """
        years = 0
        for k, v in self.factors.items():
            years += self.value.kwds.get(k, 0)/v
        return years

    @property
    def is_variable(self):
        if self.value.kwds.get('years',0) or self.value.kwds.get('months',0):
            return True
        else:
            return False

    @classmethod
    def _round_duration(cls, vals):
        """
        Round a dictionary of duration values by overflowing remainders

        The input can be
            - A numpy array of length `ss.time.factors` containing values in key order
            - A pd.DateOffset instance
            - A dictionary with keys from `ss.time.factors`

        The output will be a pd.DateOffset with integer values, where non-integer values
        have been handled by overflow using the factors in `ss.time.factors`. For example, 2.5 weeks
        would first become 2 weeks and 0.5*7 = 3.5 days, and then become 3 days + 0.5*24 = 12 hours.

        Negative values are supported - -1.5 weeks for example will become (-1w, -3d, -12h)

        Returns:
            A pd.DateOffset
        """
        if isinstance(vals, np.ndarray):
            d = sc.objdict({k:v for k,v in zip(cls.factor_keys, vals)})
        elif isinstance(vals, pd.DateOffset):
            d = sc.objdict.fromkeys(cls.factor_keys, 0)
            d.update(vals.kwds)
        elif isinstance(vals, dict):
            d = sc.objdict.fromkeys(cls.factor_keys, 0)
            d.update(vals)
        else:
            raise TypeError()

        if all([isinstance(val, int) for val in d.values()]):
            if isinstance(vals, pd.DateOffset): return vals  # pd.DateOffset is immutable so this should be OK
            return pd.DateOffset(**d)

        for i in range(len(cls.factors)-1):
            remainder, div = np.modf(d[i])
            d[i] = int(div)
            d[i+1] += remainder * cls.factor_vals[i+1]/cls.factor_vals[i]
        d[-1] = round(d[-1])

        return pd.DateOffset(**d)


    def _scale(self, dateoffset, scale):
        """
        Scale a pd.DateOffset by a factor

        This function will automatically cascade remainders to finer units using `ss.time.factors` so for
        example 2.5 weeks would first become 2 weeks and 0.5*7 = 3.5 days,
        and then become 3 days + 0.5*24 = 12 hours.

        Args:
            dateoffset: A pd.DateOffset instance
            scale: A float scaling factor (must be positive)

        Returns:
            A pd.DateOffset instance scaled by the requested amount
        """
        return self._round_duration(self._as_array(dateoffset) * scale)

    def __pow__(self, other):
        raise Exception('Cannot multiply a duration by a duration')

    def __mul__(self, other):
        if isinstance(other, Rate):
            return NotImplemented # Delegate to Rate.__rmul__
        elif isinstance(other, Dur):
            raise Exception('Cannot multiply a duration by a duration')
        return self.__class__(self._scale(self.value, other))

    def __truediv__(self, other):
        if isinstance(other, DateDur):
            # We need to convert both dates into a common timebase to do the division
            # We *can* just do self.years/other.years, however this can cause numerical precision losses
            # yet at the same time is not necessary, because the common timebase depends on the durations.
            # For example, DateDur(weeks=1)/DateDur(days=1) should return 7, but we get 6.9999999 if we convert
            # both to years. Instead though, we can just convert days into weeks instead of days into years, and
            # then divide 1 week by (1/7) days.
            # return self.years/other.years
            self_array = self.to_array()
            other_array = other.to_array()
            unit = np.argmax((self_array != 0) | (other_array != 0))
            a = 0
            b = 0
            for i, v in enumerate(self.factor_vals):
                if i < unit:
                    continue
                a += self_array[i] / v
                b += other_array[i] / v
            return a/b
        elif isinstance(other, Dur):
            return self.years/other.years
        elif isinstance(other, Rate):
            raise Exception('Cannot divide a duration by a rate')
        return self.__class__(self._scale(self.value, 1/other))

    def __repr__(self):
        if self.years == 0:
            return '<DateDur: 0>'
        else:
            labels = self.factor_keys
            vals = self.to_array().astype(float)

            time_portion = vals[4:]
            time_str = ':'.join(f'{np.round(v,1):02g}' for v in time_portion[:3])

            return '<DateDur: ' +  ','.join([f'{k}={int(v)}' for k, v in zip(labels[:4], vals[:4]) if v!=0]) + (f', +{time_str}' if time_str != '00:00:00' else '') + '>'

    def str(self):
        # Friendly representation e.g., 'day', '1 year, 2 months, 1 day'
        vals = self.to_array()

        # If we have only one nonzero value, return 'year', 'month', etc.
        if np.count_nonzero(vals == 1) == 1:
            for unit, val in zip(self.factor_keys, vals):
                if val == 1:
                    return unit[:-1]

        strs = [f'{v} {k[:-1] if abs(v) == 1 else k}' for k, v in zip(self.factor_keys, vals) if v != 0]
        return ', '.join(strs)

    def __add__(self, other):
        if isinstance(other, date):
            return other + self.value
        elif isinstance(other, DateDur):
            return self.__class__(**self._as_args(self.to_array() + self._as_array(other.value)))
        elif isinstance(other, pd.DateOffset):
            return self.__class__(**self._as_args(self.to_array() + self._as_array(other)))
        elif isinstance(other, years):
            kwargs = {k: v for k, v in zip(self.factor_keys, self.to_array())}
            kwargs['years'] += other.years
            return self.__class__(**kwargs)
        elif isinstance(other, DateArray):
            return DateArray(np.vectorize(self.__add__)(other))
        else:
            errormsg = f'For DateDur, it is only possible to add/subtract dates, Dur objects, or pd.DateOffset objects, not {type(other)}'
            raise TypeError(errormsg)

    def __sub__(self, other):
        return self.__add__(-1*other)

    def __abs__(self):
        # Cannot implement this because it's ambiguous how to resolve cases like
        # DateDur(months=-1,days=1) - this is a sensible DateDur interpreted as 'go back 1 month, then go forward 1 day'
        # but just taking the absolute value of all of the components wouldn't work because this would be on average 1 month + 1 day
        # whereas it should be 1 month - 1 day. This could probably be resolved? But is an edge case, unlikely to be needed
        # (whereas abs(years) arises when comparing dates, which automatically return years)
        raise NotImplementedError('The absolute value of a DateDur instance is undefined as components (e.g., months, days) may have different signs.')


class Rate(TimePar):
    """
    Store a value per unit time e.g., 2 per day
    - self.value - the numerator (e.g., 2) - a scalar float
    - self.unit - the denominator (e.g., 1 day) - a Dur object
    """
    base = None
    timepar_type = 'rate'
    timepar_subtype = 'rate'

    def __init__(self, value, unit=None):
        if unit is not None and self.base is not None:
            unitstr = unit.base if isinstance(unit, ss.TimePar) else unit
            if unitstr != self.base:
                errormsg = f'Cannot specify incompatible unit "{unit}" for a rate when base "{self.base}" is already specified'
                raise ValueError(errormsg)

        if unit is None:
            if self.base is not None:
                dur = class_map.dur[self.base]
                self.unit = dur(1)
            else:
                self.unit = years(1)
        elif isinstance(unit, Dur):
            self.unit = unit
        elif isinstance(unit, str):
            dur = class_map.dur[unit]
            self.unit = dur(1)
        else: # e.g. number
            self.unit = ss.years(unit) # Default of years

        if not (sc.isnumber(value) or isinstance(value, np.ndarray)):
            errormsg = f'Value must be a scalar number or array, not {type(value)}'
            raise TypeError(errormsg)
        self.value = value
        self._set_base()
        return

    def _set_base(self):
        """ Rates can have either unit or base set; handle either """
        if self.base is None and isinstance(self.unit, ss.TimePar):
            self.base = self.unit.base
        elif self.base != self.unit.base: # Should not be possible, but check just in case
            errormsg = f'Inconsistent definition: base = {self.base} but unit = {self.unit.base}'
            raise ValueError(errormsg)
        return

    def __repr__(self):
        name = self.__class__.__name__
        if name == 'Rate':
            return f'{name}({self.value}/{self.unit})'
        else:
            return f'{name}({self.value})'

    def __float__(self):
        return float(self.value)

    def __mul__(self, other):
        if isinstance(other, np.ndarray):
            return self*other
        elif isinstance(other, Dur):
            return self.value*other/self.unit
        else:
            return self.__class__(self.value*other, self.unit)

    def __neg__(self):
        return -1*self

    def __add__(self, other):
        if isinstance(other, Rate):
            if self.timepar_subtype == other.timepar_subtype:
                print('hiii', self.value, self.unit, other.value, other.unit)
                return self.__class__(self.value+other*self.unit, self.unit)
            else:
                errormsg = f'Can only add rates with the same subtype (e.g., Rate+Rate, TimeProb+TimeProb); you added {self} + {other}'
                raise TypeError(errormsg)
        else:
            if sc.isnumber(other) or isinstance(other, np.ndarray):
                errormsg = 'Only rates can be added to rates, not {other}. This error most commonly occurs if the rate needs to be multiplied by `self.dt` to get a number of events per timestep.'
                raise TypeError(errormsg)
            else:
                errormsg = 'Only rates can be added to rates, not {other}'
                raise TypeError(errormsg)

    def __radd__(self, other): return self.__add__(other)

    def __sub__(self, other):
        if self.__class__ == other.__class__:
            return self.__class__(self.value-other*self.unit, self.unit)
        elif not isinstance(other, Rate):
            if sc.isnumber(other) or isinstance(other, np.ndarray):
                raise TypeError('Only rates can be added to rates. This error most commonly occurs if the rate needs to be multiplied by `self.t.dt` to get a number of events per timestep.')
            else:
                raise TypeError('Only rates can be added to rates')
        else:
            raise TypeError('Can only subtract rates of identical types (e.g., Rate+Rate, TimeProb+TimeProb)')

    def __eq__(self, other):
        return self.value == other.value/other.unit*self.unit

    def __rmul__(self, other): return self.__mul__(other)

    def __truediv__(self, other):
        # This is for <rate>/<other>
        if isinstance(other, Rate):
            # Convert the other rate onto our dt, then divide the value
            return self.value/(other.value/other.unit*self.unit)
        elif isinstance(other, Dur):
            raise Exception('Cannot divide a rate by a duration')
        else:
            return self.__class__(self.value/other, self.unit)

    def __rtruediv__(self, other):
        # If a float is divided by a rate
        return other * (self.unit/self.value)

    def to_prob(self, dur, v=1):
        """
        Calculate a time-specific probability value

        This function is mainly useful for subclasses where the multiplication by a duration is non-linear
        (e.g., `TimeProb`) and therefore it is important to apply the factor prior to multiplication by duration.
        This function avoids creating an intermediate array of rates, and is therefore much higher performance.

        e.g.

        >>> p = ss.TimeProb(0.05)*self.cd4*self.t.dt

        and

        >>> p = ss.TimeProb(0.05).scale(self.cd4,self.t.dt)

        are equivalent, except that the second one is (much) faster.

        Args:
            v: The factor to scale the rate by. This factor is applied before multiplication by the duration
            dur: An ss.Dur instance to scale the rate by (often this is `dt`)

        Returns:
            A numpy float array of values
        """
        if isinstance(dur, np.ndarray):
            factor = (dur/self.unit).astype(float)
        else:
            factor = dur/self.unit
        return (self.value*v)*factor


class TimeProb(Rate):
    """
    `TimeProb` represents the probability of an event occurring during a
    specified period of time.

    The class is designed to allow conversion of a probability from one
    duration to another through multiplication. However, the behavior of this
    conversion depends on the data type of the the object being multiplied.

    When multiplied by a duration (type ss.Dur), the underlying constant rate is
    calculated as
        `rate = -np.log(1 - self.value)`.
    Then, the probability over the new duration is
        `p = 1 - np.exp(-rate/factor)`,
    where `factor` is the ratio of the new duration to the original duration.

    For example,
    >>> p = ss.TimeProb(0.8, ss.years(1))
    indicates a 80% chance of an event occurring in one year.

    >>> p*ss.years(1)
    When multiplied by the original denominator, 1 year in this case, the
    probability remains unchanged, 80%.

    >>> p * ss.years(2)
    Multiplying `p` by `ss.years(2)` does not simply double the
    probability to 160% (which is not possible), but rather returns a new
    probability of 96% representing the chance of the event occurring at least
    once over the new duration of two years.

    However, the behavior is different when a `TimeProb` object is multiplied
    by a scalar or array. In this case, the probability is simply scaled. This scaling
    may result in a value greater than 1, which is not valid. For example,
    >>> p * 2
    raises an AssertionError because the resulting probability (160%) exceeds 100%.

    Use `RateProb` instead if `TimeProb` if you would prefer to directly
    specify the instantaneous rate.
    """
    base = None # Can inherit, but clearer to specify
    timepar_type = 'rate'
    timepar_subtype = 'timeprob'

    def __init__(self, value, unit=None):
        try:
            assert 0 <= value <= 1, 'Value must be between 0 and 1'
        except Exception: # Something went wrong, let's figure it out -- either a value error, or value is an array instead of a scalar
            if sc.isnumber(value):
                valstr = f'Value provided: {value}'
                lt0 = value < 0
                gt1 = value > 1
            else:
                valstr = f'Values provided:\n{value}'
                lt0 = np.any(value<0)
                gt1 = np.any(value>1)
            if lt0:
                errormsg = f'Negative values are not permitted for rates or probabilities. {valstr}'
            elif gt1:
                if self.base is not None:
                    correct = f'ss.rateper{self.base}()'
                    self.__class__ = class_map.rate[f'rateper{self.base}'] # Mutate the class to the correct class
                    if ss.options.warn_convert:
                        warnmsg = f'Probabilities cannot be greater than one, so converting to a rate. Please use {correct} instead, or set ss.options.warn_convert=false. {valstr}'
                        ss.warn(warnmsg)
                else:
                    errormsg = 'Probabilities are >1, and no base was unit provided. Please use e.g. ss.rateperyear() instead of doing whatever you did. {valstr}'
                    raise ValueError(errormsg)
        Rate.__init__(self, value, unit) # Can't use super() since potentially mutating the class
        return

    def __mul__(self, other):
        if isinstance(other, np.ndarray):
            self.array_mul_error()
        elif isinstance(other, Dur):
            if self.value == 0:
                return 0
            elif self.value == 1:
                return 1
            else:
                factor = self.unit/other
                if factor == 1:
                    return self.value # Avoid expensive calculation and precision issues
                rate = -np.log(1 - self.value)
                return 1 - np.exp(-rate/factor)
        else:
            return self.__class__(self.value*other, self.unit)

    def to_prob(self, dur, v=1):
        if isinstance(dur, np.ndarray):
            factor = (dur/self.unit).astype(float)
        else:
            factor = dur/self.unit
        rate = -np.log(1-(self.value*v))
        return 1-np.exp(-rate*factor)

    @classmethod
    def array_to_prob(cls, arr, dur, v=1):
        arr = sc.promotetoarray(arr)

        if isinstance(dur, np.ndarray):
            assert arr.shape == dur.shape, 'dur must be either a scalar, or the same size as arr'
            def to_prob(a, b, _v=v):
                return a.to_prob(dur=b, v=_v)

            vectorize = np.vectorize(to_prob)
            return vectorize(arr, dur)

        elif isinstance(arr[0], TimeProb):
            factor = np.array([dur / a.unit for a in arr])
            scaled_vals = np.array([a.value * v for a in arr])
            rate = - np.log(1 - scaled_vals)
            return 1-np.exp(-rate*factor)

        else: # Assume arr is an array of values, that would be the values of a TimeProb with unit=ss.years(1)
            factor = dur / ss.years(1)
            scaled_vals = arr * v
            rate = - np.log(1 - scaled_vals)
            return 1 - np.exp(-rate * factor)

    def __truediv__(self, other): raise NotImplementedError()
    def __rtruediv__(self, other): raise NotImplementedError()


class RateProb(Rate):
    """
    A `RateProb` represents an instantaneous rate of an event occurring. Rates
    must be non-negative, but need not be less than 1.

    Through multiplication, rate can be modified or converted to a probability,
    depending on the data type of the object being multiplied.

    When a `RateProb` is multiplied by a scalar or array, the rate is simply
    scaled. Such multiplication occurs frequently in epidemiological models,
    where the base rate is multiplied by "rate ratio" or "relative rate" to
    represent agents experiencing higher (multiplier > 1) or lower (multiplier <
    1) event rates.

    Alternatively, when a `RateProb` is multiplied by a duration (type
    ss.Dur), a probability is calculated. The conversion from rate to
    probability on multiplication by a duration is
        `1 - np.exp(-rate/factor)`,
    where `factor` is the ratio of the multiplied duration to the original
    period (denominator).

    For example, consider
    >>> p = ss.RateProb(0.8, ss.years(1))
    When multiplied by a duration of 1 year, the calculated probability is
        `1 - np.exp(-0.8)`, which is approximately 55%.
    >>> p*ss.years(1)

    When multiplied by a scalar, the rate is simply scaled.
    >>> p*2

    The difference between `TimeProb` and `RateProb` is subtle, but important. `RateProb` works directly
    with the instantaneous rate of an event occurring. In contrast, `TimeProb` starts with a probability and a duration,
    and the underlying rate is calculated. On multiplication by a duration,
    * RateProb: rate -> probability
    * TimeProb: probability -> rate -> probability

    The behavior of both classes is depending on the data type of the object being multiplied.
    """
    base = None # Can inherit, but clearer to specify
    timepar_type = 'rate'
    timepar_subtype = 'rateprob'

    def __init__(self, value, unit=None):
        assert value >= 0, 'Value must be >= 0'
        return super().__init__(value, unit)

    def __mul__(self, other):
        if isinstance(other, np.ndarray):
            self.array_mul_error()
        elif isinstance(other, Dur):
            if self.value == 0:
                return 0
            else:
                factor = self.unit/other
                return 1 - np.exp(-self.value/factor)
        else:
            return self.__class__(self.value*other, self.unit)

    def to_prob(self, dur, v=1):
        if isinstance(dur, np.ndarray):
            factor = (dur/self.unit).astype(float)
        else:
            factor = dur/self.unit
        rate = self.value*v
        return 1-np.exp(-rate*factor)

    def __truediv__(self, other):
        if isinstance(other, Dur):
            raise NotImplementedError()
        return super().__truediv__(other)  # Fixed the call to super().__truediv__

    def __rtruediv__(self, other): raise NotImplementedError()


#%% Convenience classes
__all__ += ['years', 'months', 'weeks', 'days', 'year', 'month', 'week', 'day', # Durations
            'perday', 'perweek', 'permonth', 'peryear', # TimeProbs
            'probperday', 'probperweek', 'probpermonth', 'probperyear', # TimeProb aliases
            'rateperday', 'rateperweek', 'ratepermonth', 'rateperyear'] # Rates

# Durations
class years(Dur):  base = 'year'
class months(Dur): base = 'month'
class weeks(Dur):  base = 'week'
class days(Dur):   base = 'day'

# Duration shortcuts
year = years(1)
month = months(1)
week = weeks(1)
day = days(1)
for obj in [year, month, week, day]:
    object.__setattr__(obj, '_locked', True) # Make immutable

# TimeProbs
class perday(TimeProb):   base = 'day'
class perweek(TimeProb):  base = 'week'
class permonth(TimeProb): base = 'month'
class peryear(TimeProb):  base = 'year'

# Aliases
probperday = perday
probperweek = perweek
probpermonth = permonth
probperyear = peryear

# Rates
class rateperday(Rate):   base = 'day'
class rateperweek(Rate):  base = 'week'
class ratepermonth(Rate): base = 'month'
class rateperyear(Rate):  base = 'year'

# Define a mapping of all the options -- dictionary lookup is O(1) so it's OK to have a lot of keys
reverse_class_map = {
    days:   ['d', 'day', 'days', day, days],
    weeks:  ['w', 'week', 'weeks', week, weeks],
    months: ['m', 'month', 'months', month, months],
    years:  ['y', 'year', 'years', year, years],

    perday:   ['perday', 'probperday', perday],
    perweek:  ['perweek', 'probperweek', perweek],
    permonth: ['permonth', 'probpermonth', permonth],
    peryear:  ['peryear', 'probperyear', peryear],

    rateperday:   ['rateperday', rateperday],
    rateperweek:  ['rateperweek', rateperweek],
    ratepermonth: ['ratepermonth', ratepermonth],
    rateperyear:  ['rateperyear', rateperyear],
}

# Convert to the actual class map
class_map = sc.objdict(dur=sc.objdict(), rate=sc.objdict()) # TODO: make a class with an informative error message for invalid lookups (e.g. dur vs rate)
for v, keylist in reverse_class_map.items():
    for key in keylist:
        if issubclass(v, Dur):
            class_map.dur[key] = v
        elif issubclass(v, Rate):
            class_map.rate[key] = v
        else:
            errormsg = f'Unexpected entry: {v}'
            raise TypeError(errormsg)


"""
Experimenting with a sexual orentation class
"""

import starsim as ss
import numpy as np

#########################################################
# Method using only 1 sexual network and defining sexual orientation
#########################################################
class orientation:
    def __init__(self, options=None, pop_shares=None, m_partner_prob=None):
        self.options = options
        self.pop_shares = pop_shares
        self.m_partner_prob = m_partner_prob


# Motivating examples:
# 1. Sexual orientation patterns from Western countries
western_prefs = orientation(
    options=[0, 1, 2, 3, 4],  # Abridged Kinsey scale
    m_partner_prob=[0.00, 0.005, 0.020, 0.100, 1.00],  # Probability of taking a male partner
    pop_shares=[0.93, 0.040, 0.005, 0.005, 0.02]
    # From https://en.wikipedia.org/wiki/Demographics_of_sexual_orientation
)
orientation1 = ss.State('orientation', int,
                           distdict=dict(dist='choice', par1=western_prefs.m_partner_prob,
                                         par2=western_prefs.pop_shares),
                           eligibility='male')

# 2. Sexual orientation with no MSM/MF mixing
exclusive_prefs = orientation(
    options=['mf', 'msm'],  # No mixing
    m_partner_prob=[0.0, 1.0],  # Probability of taking a male partner
    pop_shares=[0.95, 0.05]
)

# 3. Sexual orientation without open MSM
closeted_prefs = orientation(
    options=[0, 1, 2, 3],  # last category still only partners with MSM 50% of the time
    m_partner_prob=[0.0, 0.005, 0.02, 0.50],  # Probability of taking a male partner
    pop_shares=[0.9, 0.010, 0.03, 0.06]
)

#########################################################
# Multidimensional preferences
#########################################################
# Make a person with an age, sex, geolocation, and preferences for all these attributes in their partner
class Person:
    def __init__(self, name, sex, exact_age, geo, sex_prefs=None, age_prefs=None, geo_prefs=None):
        self.name = name
        self.sex = sex
        self.exact_age = exact_age
        self.age_bins = np.array([15,30,45,60,100])
        self.age = self.age_bins[np.digitize(self.exact_age, self.age_bins)]
        self.geo = geo
        self.prefs = {'sex': sex_prefs, 'age': age_prefs, 'geo': geo_prefs}
        self.pop_rankings = dict()
        self.pop_scores = None
        self.pop_order = None

    def get_preferences(self, potentials):
        # Example ranking algorithm
        self.pop_scores = np.ones_like(potentials)

        for this_dim, these_prefs in p.prefs.items():
            prefs = np.zeros_like(potentials)
            partner_vals = np.array([getattr(q, this_dim) for q in potentials])
            for s, sprob in these_prefs.items():
                matches = ss.true(partner_vals == s)
                if len(matches):
                    prefs[matches] = sprob / len(matches)
            self.pop_rankings[this_dim] = prefs
            self.pop_scores = self.pop_scores * prefs

        self.pop_order = np.array([q.name for q in potentials])[np.argsort(self.pop_scores)][::-1]
        # self.pop_order = list(self.pop_order)


if __name__ == '__main__':
    # Run tests

    p1 = Person('p1', 1, 22, 'A', sex_prefs={0:0.8, 1:0.2}, age_prefs={15:0, 30:0.99, 45:0.01, 60:0, 100:0}, geo_prefs={'A':0.6, 'B':0.3, 'C':0.1})
    p2 = Person('p2', 0, 29, 'B', sex_prefs={0:0.0, 1:1.0}, age_prefs={15:0, 30:0.5, 45:0.5, 60:0, 100:0}, geo_prefs={'A':0.2, 'B':0.7, 'C':0.1})
    p3 = Person('p3', 0, 40, 'A', sex_prefs={0:0.0, 1:1.0}, age_prefs={15:0, 30:0.4, 45:0.6, 60:0, 100:0}, geo_prefs={'A':0.9, 'B':0.1, 'C':0.0})
    p4 = Person('p4', 1, 20, 'A', sex_prefs={0:0.02, 1:0.98}, age_prefs={15:0, 30:0.9, 45:0.1, 60:0, 100:0}, geo_prefs={'A':0.5, 'B':0.5, 'C':0.0})

    population = [p1, p2, p3, p4]
    for p in population:
        potentials = [q for q in population if q != p]
        p.get_preferences(potentials)

# How does each person rank each person??
print(f'p1: {p1.pop_order}\n')
print(f'p2: {p2.pop_order}\n')

# These are not quite right...

# Pairs:
# [p1, p2], [p3, p4]
# [p1, p4], [na, na]



"""
Numerical utilities
"""
import warnings
import numpy as np
import numba as nb
import pandas as pd
import sciris as sc
import matplotlib.pyplot as plt
import starsim as ss

# %% Helper functions

# What functions are externally visible
__all__ = ['ndict', 'warn', 'find_contacts', 'set_seed', 'check_requires', 'standardize_netkey',
           'standardize_data', 'validate_sim_data', 'load', 'save', 'return_fig']


class ndict(sc.objdict):
    """
    A dictionary-like class that provides additional functionalities for handling named items.

    Args:
        name (str): The attribute of the item to use as the dict key (i.e., all items should have this attribute defined)
        type (type): The expected type of items.
        strict (bool): If True, only items with the specified attribute will be accepted.
        overwrite (bool): whether to allow adding a key when one has already been added

    **Examples**::

        networks = ss.ndict(ss.MFNet(), ss.MaternalNet())
        networks = ss.ndict([ss.MFNet(), ss.MaternalNet()])
        networks = ss.ndict({'mf':ss.MFNet(), 'maternal':ss.MaternalNet()})
    """

    def __init__(self, *args, nameattr='name', type=None, strict=True, overwrite=False, **kwargs):
        super().__init__()
        self.setattribute('_nameattr', nameattr)  # Since otherwise treated as keys
        self.setattribute('_type', type)
        self.setattribute('_strict', strict)
        self.setattribute('_overwrite', overwrite)
        self.extend(*args, **kwargs)
        return

    def __call__(self):
        """ Shortcut for returning values """
        return self.values()

    def append(self, arg, key=None, overwrite=None):
        valid = False
        if arg is None:
            return  # Nothing to do
        elif hasattr(arg, self._nameattr):
            key = key or getattr(arg, self._nameattr)
            valid = True
        elif isinstance(arg, dict):
            if self._nameattr in arg:
                key = key or arg[self._nameattr]
                valid = True
            else:
                for k, v in arg.items():
                    self.append(v, key=k)
                valid = None  # Skip final processing
        elif not self._strict:
            key = key or f'item{len(self) + 1}'
            valid = True
        else:
            valid = False

        if valid is True:
            if self._strict:
                self._check_type(arg)
                self._check_key(key, overwrite=overwrite)
            self[key] = arg # Actually add to the ndict!
        elif valid is None:
            pass  # Nothing to do
        else:
            errormsg = f'Could not interpret argument {arg}: does not have expected attribute "{self._nameattr}"'
            raise TypeError(errormsg)
        return self

    def _check_key(self, key, overwrite=None):
        if overwrite is None: overwrite = self._overwrite
        if key in self:
            if not overwrite:
                typestr = f' "{self._type}"' if self._type else ''
                errormsg = f'Cannot add object "{key}" since already present in ndict{typestr} with keys:\n{sc.newlinejoin(self.keys())}'
                raise ValueError(errormsg)
            else:
                ss.warn(f'Overwriting existing ndict entry "{key}"')
        return

    def _check_type(self, arg):
        """ Check types """
        if self._type is not None:
            if not isinstance(arg, self._type) and not isinstance(arg, ss.Module): # Module is a valid argument anywhere
                errormsg = f'The following item does not have the expected type {self._type}:\n{arg}'
                raise TypeError(errormsg)
        return

    def extend(self, *args, **kwargs):
        """ Add new items to the ndict, by item, list, or dict """
        args = sc.mergelists(*args)
        for arg in args:
            self.append(arg)
        for key, arg in kwargs.items():
            self.append(arg, key=key)
        return self

    def merge(self, other):
        """ Merge another dictionary with this one """
        for key, arg in other.items():
            self.append(arg, key=key)
        return self

    def copy(self):
        """ Shallow copy """
        new = self.__class__.__new__(nameattr=self._nameattr, type=self._type, strict=self._strict)
        new.update(self)
        return new

    def __add__(self, other):
        """ Allow c = a + b """
        new = self.copy()
        if isinstance(other, list):
            new.extend(other)
        else:
            new.append(other)
        return new

    def __iadd__(self, other):
        """ Allow a += b """
        if isinstance(other, list):
            self.extend(other)
        else:
            self.append(other)
        return self


def warn(msg, category=None, verbose=None, die=None):
    """ Helper function to handle warnings -- shortcut to warnings.warn """

    # Handle inputs
    warnopt = ss.options.warnings if not die else 'error'
    if category is None:
        category = RuntimeWarning
    if verbose is None:
        verbose = ss.options.verbose

    # Handle the different options
    if warnopt in ['error', 'errors']:  # Include alias since hard to remember
        raise category(msg)
    elif warnopt == 'warn':
        msg = '\n' + msg
        warnings.warn(msg, category=category, stacklevel=2)
    elif warnopt == 'print':
        if verbose:
            msg = 'Warning: ' + msg
            print(msg)
    elif warnopt == 'ignore':
        pass
    else:
        options = ['error', 'warn', 'print', 'ignore']
        errormsg = f'Could not understand "{warnopt}": should be one of {options}'
        raise ValueError(errormsg)

    return


def find_contacts(p1, p2, inds):  # pragma: no cover
    """
    Variation on Network.find_contacts() that avoids sorting.

    A set is returned here rather than a sorted array so that custom tracing interventions can efficiently
    add extra people. For a version with sorting by default, see Network.find_contacts(). Indices must be
    an int64 array since this is what's returned by true() etc. functions by default.
    """
    pairing_partners = set()
    inds = set(inds)
    for i in range(len(p1)):
        if p1[i] in inds:
            pairing_partners.add(p2[i])
        if p2[i] in inds:
            pairing_partners.add(p1[i])
    return pairing_partners


def check_requires(sim, requires, *args):
    """ Check that the module's requirements (of other modules) are met """
    errs = sc.autolist()
    all_classes = [m.__class__ for m in sim.modules]
    all_names = [m.name for m in sim.modules]
    for req in sc.mergelists(requires, *args):
        if req not in all_classes + all_names:
            errs += req
    if len(errs):
        errormsg = f'The following module(s) are required, but the Sim does not contain them: {sc.strjoin(errs)}'
        raise AttributeError(errormsg)
    return


def set_seed(seed=None):
    '''
    Reset the random seed -- complicated because of Numba, which requires special
    syntax to reset the seed. This function also resets Python's built-in random
    number generated.

    Args:
        seed (int): the random seed
    '''

    @nb.njit(cache=True)
    def set_seed_numba(seed):
        return np.random.seed(seed)

    def set_seed_regular(seed):
        return np.random.seed(seed)

    # Dies if a float is given
    if seed is not None:
        seed = int(seed)

    set_seed_regular(seed)  # If None, reinitializes it
    if seed is None:  # Numba can't accept a None seed, so use our just-reinitialized Numpy stream to generate one
        seed = np.random.randint(1e9)
    set_seed_numba(seed)

    return


# %% Data cleaning and processing

def standardize_netkey(key):
    """ Networks can be upper or lowercase, and have a suffix 'net' or not; this function standardizes them """
    return key.lower().removesuffix('net')


def standardize_data(data=None, metadata=None, min_year=1800, out_of_range=0, default_age=0, default_year=2024):
    """
    Standardize formats of input data

    Input data can arrive in many different forms. This function accepts a variety of data
    structures, and converts them into a Pandas Series containing one variable, based on
    specified metadata, or an ``ss.Dist`` if the data is already an ``ss.Dist`` object.

    The metadata is a dictionary that defines columns of the dataframe or keys
    of the dictionary to use as indices in the output Series. It should contain:

    - ``metadata['data_cols']['value']`` specifying the name of the column/key to draw values from
    - ``metadata['data_cols']['year']`` optionally specifying the column containing year values; otherwise the default year will be used
    - ``metadata['data_cols']['age']`` optionally specifying the column containing age values; otherwise the default age will be used
    - ``metadata['data_cols'][<arbitrary>]`` optionally specifying any other columns to use as indices. These will form part of the multi-index for the standardized Series output.

    If a ``sex`` column is part of the index, the metadata can also optionally specify a string mapping to convert
    the sex labels in the input data into the 'm'/'f' labels used by Starsim. In that case, the metadata can contain
    an additional key like ``metadata['sex_keys'] = {'Female':'f','Male':'m'}`` which in this case would map the strings
    'Female' and 'Male' in the original data into 'm'/'f' for Starsim.

    Args:
        data (pandas.DataFrame, pandas.Series, dict, int, float): An associative array  or a number, with the input data to be standardized.
        metadata (dict): Dictionary specifiying index columns, the value column, and optionally mapping for sex labels
        min_year (float): Optionally specify a minimum year allowed in the data. Default is 1800.
        out_of_range (float): Value to use for negative ages - typically 0 is a reasonable choice but other values (e.g., np.inf or np.nan) may be useful depending on the calculation. This will automatically be added to the dataframe with an age of ``-np.inf``

    Returns:

        - A `pd.Series` for all supported formats of `data` *except* an ``ss.Dist``. This series will contain index columns for 'year'
          and 'age' (in that order) and then subsequent index columns for any other variables specified in the metadata, in the order
          they appeared in the metadata (except for year and age appearing first).
        - An ``ss.Dist`` instance - if the ``data`` input is an ``ss.Dist``, that same object will be returned by this function
    """
    # It's a format that can be used directly: return immediately
    if sc.isnumber(data) or isinstance(data, (ss.Dist, ss.TimePar)):
        return data

    # Convert series and dataframe inputs into dicts
    if isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        data = data.reset_index().to_dict(orient='list')

    # Check that the input is now a dict (scalar types have already been handled above
    assert isinstance(data, dict), 'Supported inputs are ss.Dict, scalar numbers, DataFrames, Series, or dictionaries'

    # Extract the values and index columns
    assert 'value' in metadata['data_cols'], 'The metadata is missing a column name for "value", which must be provided if the input data is in the form of a DataFrame, Series, or dict'
    values = sc.promotetoarray(data[metadata['data_cols']['value']])
    index = sc.objdict()
    for k, col in metadata['data_cols'].items():
        if k != 'value':
            index[k] = data[col]

    # Add defaults for year and age
    if 'year' not in index:
        index['year'] = np.full(values.shape, fill_value=default_year, dtype=ss.dtypes.float)
    if 'age' not in index:
        index['age'] = np.full(values.shape, fill_value=default_age, dtype=ss.dtypes.float)

    # Reorder the index so that it starts with age first
    index.insert(0, 'age', index.pop('age'))
    index.insert(0, 'year', index.pop('year'))

    # Map sex values
    if 'sex' in index:
        if 'sex_keys' in metadata:
            index['sex'] = [metadata['sex_keys'][x] for x in index['sex']]
        assert set(index['sex']) == {'f','m'}, 'If standardized data contains a "sex" column, it should use "m" and "f" to specify the sex.'

    # Create the series
    output = pd.Series(data=values, index=pd.MultiIndex.from_arrays(index.values(), names=index.keys()))

    # Add an entry for negative ages
    new_entries = output.index.set_codes([0] * len(output), level=1).set_levels([-np.inf], level=1).unique()
    new = pd.Series(data=out_of_range, index=new_entries)
    output = pd.concat([output, new])

    # Truncate years
    output = output.loc[output.index.get_level_values('year')>min_year]

    # Perform a final sort to prevent "indexing past lexsort depth" warnings
    output = output.sort_index()

    return output


def validate_sim_data(data=None, die=None):
    """
    Validate data intended to be compared to the sim outputs, e.g. for calibration

    Args:
        data (df/dict): a dataframe (or dict) of data, with a column "time" plus data columns of the form "module.result", e.g. "hiv.new_infections"
        die (bool): whether to raise an exception if the data cannot be converted (default: die if data is not None but cannot be converted)

    """
    success = False
    if data is not None:
        # Try loading the data
        try:
            data = sc.dataframe(data) # Convert it to a dataframe
            timecols = ['t', 'timevec', 'tvec', 'time', 'day', 'date', 'year'] # If a time column is supplied, use it as the index
            found = False
            for timecol in timecols:
                if timecol in data.cols:
                    if found:
                        errormsg = f'Multiple time columns found: please ensure only one of {timecols} is present; you supplied {data.cols}.'
                        raise ValueError(errormsg)
                    data.set_index(timecol, inplace=True)
                    found = True
            success = True

        # Data loading failed
        except Exception as E:
            errormsg = f'Failed to add data "{data}": expecting a dataframe-compatible object. Error:\n{E}'
            if die == False:
                print(errormsg)
            else:
                raise ValueError(errormsg)

    # Validation
    if not success and die == True:
        errormsg = 'Data "{data}" could not be converted and die == True'
        raise ValueError(errormsg)

    return data


def combine_rands(a, b):
    """
    Efficient algorithm for combining two arrays of random integers into an array
    of floats.

    See ss.multi_random() for the user-facing version.

    Args:
        a (array): array of random integers between 0 and np.iinfo(np.uint64).max, as from ss.rand_raw()
        b (array): ditto, same size as a

    Returns:
        A new array of random numbers the same size as a and b
    """
    c = np.bitwise_xor(a*b, a-b)
    u = c / np.iinfo(np.uint64).max
    return u


#%% Other helper functions

def load(filename, **kwargs):
    """
    Alias to Sciris sc.loadany()

    Since Starsim uses Sciris for saving objects, they can be loaded back using
    this function. This can also be used to load other objects of known type
    (e.g. JSON), although this usage is discouraged.

    Args:
        filename (str/path): the name of the file to load
        kwargs (dict): passed to sc.loadany()

    Returns:
        The loaded object
    """
    return sc.loadany(filename, **kwargs)


def save(filename, obj, **kwargs):
    """
    Alias to Sciris sc.save()

    While some Starsim objects have their own save methods, this function can be
    used to save any arbitrary object. It can then be loaded with ss.load().

    Args:
        filename (str/path): the name of the file to save
        obj (any): the object to save
        kwargs (dict): passed to sc.save()
    """
    return sc.save(filename=filename, obj=obj, **kwargs)


class shrink:
    """ Define a class to indicate an object has been shrunken """
    def __repr__(self):
        s = 'This object has been intentionally "shrunken"; it is a placeholder and has no functionality. Use the non-shrunken object instead.'
        return s


def return_fig(fig, **kwargs):
    """ Do postprocessing on the figure: by default, don't return if in Jupyter, but show instead """
    is_jupyter = [False, True, sc.isjupyter()][ss.options.jupyter]
    is_reticulate = ss.options.reticulate
    if is_jupyter or is_reticulate:
        print(fig)
        plt.show()
        return None
    else:
        return fig


import pandas as pd
import numpy as np
import sciris as sc
from starsim.states import State, ArrayView

INT_NAN = np.iinfo(int).max  # Value to use to flag removed UIDs (i.e., an integer value we are treating like NaN, since NaN can't be stored in an integer array)

# Some gotchas
# - If two FusedArrays have the same number of entries but have different UIDs (or even a different order of UIDs) then operations like addition will work
#   but they will produce incorrect output. States all reference the same UIDs so can be safely operated on


class DynamicPeople():
    def __init__(self, states=None):

        self._uid_map = ArrayView(int, fill_value=INT_NAN) # This variable tracks all UIDs ever created
        self.uid = ArrayView(int, fill_value=INT_NAN) # This variable tracks all UIDs currently in use

        self.sex = State('sex',bool)
        self.dead = State('dead',bool)

        self.states = [self.sex, self.dead] # All state objects linked to this People instance. Note that the states above are not yet linked to the People because they haven't been initialized

        # In reality, these states might be added elsewhere (e.g., in modules)
        if states is not None:
            for state in states:
                self.add_state(state)

    def add_state(self, state):
        if id(state) not in {id(x) for x in self.states}:
            self.states.append(state)

    def __len__(self):
        return len(self.uid)

    def initialize(self, n):

        self._uid_map.grow(n)
        self._uid_map[:] = np.arange(0, n)
        self.uid.grow(n)
        self.uid[:] = np.arange(0, n)

        for state in self.states:
            state.initialize(self)

    def __setattr__(self, attr, value):
        if hasattr(self, attr) and isinstance(getattr(self, attr), State):
            raise Exception('Cannot assign directly to a state - must index into the view instead e.g. `people.uid[:]=`')
        else:   # If not initialized, rely on the default behavior
            object.__setattr__(self, attr, value)

    def grow(self, n):
        # Add agents
        start_uid = len(self._uid_map)
        start_idx = len(self.uid)

        new_uids = np.arange(start_uid, start_uid+n)
        new_inds = np.arange(start_idx, start_idx+n)

        self._uid_map.grow(n)
        self._uid_map[new_uids] = new_inds

        self.uid.grow(n)
        self.uid[new_inds] = new_uids

        for state in self.states:
            state.grow(n)

    def remove(self, uids_to_remove):
        # Calculate the *indices* to keep
        keep_uids = self.uid[~np.in1d(self.uid, uids_to_remove)] # Calculate UIDs to keep
        keep_inds = self._uid_map[keep_uids] # Calculate indices to keep

        # Trim the UIDs and states
        self.uid._trim(keep_inds)
        for state in self.states:
            state._trim(keep_inds)

        # Update the UID map
        self._uid_map[:] = INT_NAN # Clear out all previously used UIDs
        self._uid_map[keep_uids] = np.arange(0, len(keep_uids)) # Assign the array indices for all of the current UIDs

if __name__ == '__main__':
    x = State('foo', int, 0)
    z = State('bar', int, lambda n: np.random.randint(1,3,n))

    p = DynamicPeople(states=[x,z])
    p.initialize(200000)

    remove = np.random.choice(np.arange(len(p)), 100000, replace=False)
    p.remove(remove)

    # Define uids and indices
    multiple_items_uid = np.random.choice(p.uid, 50000, replace=False)
    multiple_items_ind = p._uid_map[multiple_items_uid]
    multiple_items_few_uid = np.random.choice(p.uid, 1000, replace=False)
    multiple_items_few_ind = p._uid_map[multiple_items_few_uid]
    single_item_uid = multiple_items_uid[5000]
    single_item_ind = multiple_items_ind[5000]
    boolean = np.random.randint(0,2, size=len(p)).astype(bool)

    a = z.values
    s = pd.Series(z.values, p.uid)
    uid_map = p._uid_map._view

    def lookup(lbl, uids, n=2_000):
        print('\n' + lbl)
        with sc.timer('State: '):
            for i in range(n): z[uids]
        with sc.timer('Series: '):
            for i in range(n): s[uids]
        with sc.timer('Series (loc): '):
            for i in range(n): s.loc[uids]

        with sc.timer('Series (hack): '):
            for i in range(n): 
                if uids.dtype == np.bool_:
                    s.values[uids]
                else:
                    s.values[uid_map[uids]]

    def assign(lbl, uids, vals, n=2_000):
        print('\n' + lbl)
        with sc.timer('State: '):
            for i in range(n): z[uids] = vals
        with sc.timer('Series: '):
            for i in range(n): s[uids] = vals
        with sc.timer('Series (loc): '):
            for i in range(n): s.loc[uids] = vals

        with sc.timer('Series (hack): '):
            for i in range(n): 
                if uids.dtype == np.bool_:
                    s.values[uids] = vals
                else:
                    s.values[uid_map[uids]] = vals

    lookup('SINGLE ITEM LOOKUP', single_item_uid)
    lookup('MULTIPLE ITEM LOOKUP (state lookup also includes construction of a FusedArray)', multiple_items_uid)
    lookup('MULTIPLE ITEM LOOKUP (FEW) (state lookup also includes construction of a FusedArray)', multiple_items_few_uid)
    lookup('BOOLEAN ARRAY LOOKUP', boolean)

    assign('SINGLE ITEM ASSIGNMENT', single_item_uid, 1)
    assign('MULTIPLE ITEM ASSIGNMENT', multiple_items_uid, 1)
    assign('MULTIPLE ITEM ASSIGNMENT (FEW)', multiple_items_few_uid, 1)
    
    assign('MULTIPLE ITEM ARRAY ASSIGNMENT', multiple_items_uid, multiple_items_ind)
    assign('MULTIPLE ITEM ARRAY ASSIGNMENT (FEW)', multiple_items_few_uid, multiple_items_few_ind)
    assign('BOOLEAN ARRAY ASSIGNMENT', boolean, 1)


"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '2.3.2'
__versiondate__ = '2025-07-16'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2025 by IDM'



"""
Define SIR and SIS disease modules
"""

import numpy as np
import sciris as sc
import matplotlib.pyplot as plt
import starsim as ss


__all__ = ['SIR', 'SIS']

class SIR(ss.Infection):
    """
    Example SIR model

    This class implements a basic SIR model with states for susceptible,
    infected/infectious, and recovered. It also includes deaths, and basic
    results.
    """
    def __init__(self, pars=None, **kwargs):
        super().__init__()
        self.define_pars(
            beta = ss.beta(0.1),
            init_prev = ss.bernoulli(p=0.01),
            dur_inf = ss.lognorm_ex(mean=ss.dur(6)),
            p_death = ss.bernoulli(p=0.01),
        )
        self.update_pars(pars, **kwargs)

        self.define_states(
            ss.State('susceptible', default=True, label='Susceptible'),
            ss.State('infected', label='Infectious'),
            ss.State('recovered', label='Recovered'),
            ss.FloatArr('ti_infected', label='Time of infection'),
            ss.FloatArr('ti_recovered', label='Time of recovery'),
            ss.FloatArr('ti_dead', label='Time of death'),
            ss.FloatArr('rel_sus', default=1.0, label='Relative susceptibility'),
            ss.FloatArr('rel_trans', default=1.0, label='Relative transmission'),
        )
        return

    def step_state(self):
        # Progress infectious -> recovered
        sim = self.sim
        recovered = (self.infected & (self.ti_recovered <= sim.ti)).uids
        self.infected[recovered] = False
        self.recovered[recovered] = True

        # Trigger deaths
        deaths = (self.ti_dead <= sim.ti).uids
        if len(deaths):
            sim.people.request_death(deaths)
        return

    def set_prognoses(self, uids, sources=None):
        """ Set prognoses """
        super().set_prognoses(uids, sources)
        ti = self.t.ti
        self.susceptible[uids] = False
        self.infected[uids] = True
        self.ti_infected[uids] = ti

        p = self.pars

        # Sample duration of infection, being careful to only sample from the
        # distribution once per timestep.
        dur_inf = p.dur_inf.rvs(uids)

        # Determine who dies and who recovers and when
        will_die = p.p_death.rvs(uids)
        dead_uids = uids[will_die]
        rec_uids = uids[~will_die]
        self.ti_dead[dead_uids] = ti + dur_inf[will_die] # Consider rand round, but not CRN safe
        self.ti_recovered[rec_uids] = ti + dur_inf[~will_die]
        return

    def step_die(self, uids):
        """ Reset infected/recovered flags for dead agents """
        self.susceptible[uids] = False
        self.infected[uids] = False
        self.recovered[uids] = False
        return

    def plot(self, **kwargs):
        """ Default plot for SIR model """
        fig = plt.figure()
        kw = sc.mergedicts(dict(lw=2, alpha=0.8), kwargs)
        res = self.results
        for rkey in ['n_susceptible', 'n_infected', 'n_recovered']:
            plt.plot(res.timevec, res[rkey], label=res[rkey].label, **kw)
        plt.legend(frameon=False)
        plt.xlabel('Time')
        plt.ylabel('Number of people')
        plt.ylim(bottom=0)
        sc.boxoff()
        sc.commaticks()
        return ss.return_fig(fig)


class SIS(ss.Infection):
    """
    Example SIS model

    This class implements a basic SIS model with states for susceptible,
    infected/infectious, and back to susceptible based on waning immunity. There
    is no death in this case.
    """
    def __init__(self, pars=None, *args, **kwargs):
        super().__init__()
        self.define_pars(
            beta = ss.beta(0.05),
            init_prev = ss.bernoulli(p=0.01),
            dur_inf = ss.lognorm_ex(mean=ss.dur(10)),
            waning = ss.rate(0.05),
            imm_boost = 1.0,
        )
        self.update_pars(pars=pars, *args, **kwargs)

        self.define_states(
            ss.FloatArr('ti_recovered'),
            ss.FloatArr('immunity', default=0.0),
        )
        return

    def step_state(self):
        """ Progress infectious -> recovered """
        recovered = (self.infected & (self.ti_recovered <= self.ti)).uids
        self.infected[recovered] = False
        self.susceptible[recovered] = True
        self.update_immunity()
        return

    def update_immunity(self):
        has_imm = (self.immunity > 0).uids
        self.immunity[has_imm] = (self.immunity[has_imm])*(1 - self.pars.waning)
        self.rel_sus[has_imm] = np.maximum(0, 1 - self.immunity[has_imm])
        return

    def set_prognoses(self, uids, sources=None):
        """ Set prognoses """
        super().set_prognoses(uids, sources)
        self.susceptible[uids] = False
        self.infected[uids] = True
        self.ti_infected[uids] = self.ti
        self.immunity[uids] += self.pars.imm_boost

        # Sample duration of infection
        dur_inf = self.pars.dur_inf.rvs(uids)

        # Determine when people recover
        self.ti_recovered[uids] = self.ti + dur_inf

        return

    def init_results(self):
        """ Initialize results """
        super().init_results()
        self.define_results(
            ss.Result('rel_sus', dtype=float, label='Relative susceptibility')
        )
        return

    def update_results(self):
        """ Store the population immunity (susceptibility) """
        super().update_results()
        self.results['rel_sus'][self.ti] = self.rel_sus.mean()
        return

    def plot(self, **kwargs):
        """ Default plot for SIS model """
        fig = plt.figure()
        kw = sc.mergedicts(dict(lw=2, alpha=0.8), kwargs)
        res = self.results
        for rkey in ['n_susceptible', 'n_infected']:
            plt.plot(res.timevec, res[rkey], label=res[rkey].label, **kw)
        plt.legend(frameon=False)
        plt.xlabel('Time')
        plt.ylabel('Number of people')
        plt.ylim(bottom=0)
        sc.boxoff()
        sc.commaticks()
        return ss.return_fig(fig)


# %% Interventions

__all__ += ['sir_vaccine']

class sir_vaccine(ss.Vx):
    """
    Create a vaccine product that affects the probability of infection.

    The vaccine can be either "leaky", in which everyone who receives the vaccine
    receives the same amount of protection (specified by the efficacy parameter)
    each time they are exposed to an infection. The alternative (leaky=False) is
    that the efficacy is the probability that the vaccine "takes", in which case
    that person is 100% protected (and the remaining people are 0% protected).

    Args:
        efficacy (float): efficacy of the vaccine (0<=efficacy<=1)
        leaky (bool): see above
    """
    def __init__(self, pars=None, *args, **kwargs):
        super().__init__()
        self.define_pars(
            efficacy = 0.9,
            leaky = True
        )
        self.update_pars(pars, **kwargs)
        return

    def administer(self, people, uids):
        if self.pars.leaky:
            people.sir.rel_sus[uids] *= 1-self.pars.efficacy
        else:
            people.sir.rel_sus[uids] *= np.random.binomial(1, 1-self.pars.efficacy, len(uids))
        return



"""
Simulation and module timelines
"""
import sciris as sc
import numpy as np
import starsim as ss

__all__ = ['Timeline']


class Timeline:
    """
    Handle time vectors for both simulations and modules.

    Each module can have its own time instance, in the case where the time vector
    is defined by absolute dates, these time vectors are by definition aligned. Otherwise
    they can be specified using Dur objects which express relative times (they can be added
    to a date to get an absolute time)

    Args:
        start : ss.date or ss.Dur
        stop : ss.date if start is an ss.date, or an ss.Dur if start is an ss.Dur
        dt (ss.Dur): Simulation step size
        pars (dict): if provided, populate parameter values from this dictionary
        parent (obj): if provided, populate missing parameter values from a 'parent" `Timeline` instance
        name (str): if provided, name the `Timeline` object
        init (bool): whether or not to immediately initialize the Timeline object
        sim (bool/Sim): if True, initializes as a sim-specific `Timeline` instance; if a Sim instance, initialize the absolute time vector

    The `Timeline` object, after initialization, has the following attributes:

    - `ti` (int): the current timestep
    - `npts` (int): the number of timesteps
    - `tvec` (array): time either as absolute `ss.date` instances, or relative `ss.Dur` instances
    - `yearvec` (array): time represented as floating-point years

    **Examples**:

        t1 = ss.Timeline(start=2000, stop=2020, dt=1.0)
        t2 = ss.Timeline(start='2021-01-01', stop='2021-04-04', dt=ss.days(2))
    """

    # Allowable time arguments
    time_args = ['start', 'stop', 'dt']

    def __init__(self, start=None, stop=None, dt=None, dur=None, name=None):
        # Store inputs
        self.name = name
        self.start = start
        self.stop = stop
        self.dt = dt
        self.dur = dur

        # Set defaults
        self.default_dur   = ss.years(50)
        self.default_start = ss.years(2000)
        self.default_dt    = ss.years(1.0)
        self.calendar_year_threshold = 1900.0 # Value at which to switch over from interpreting a value from a year duration to a calendar year

        # Populated later
        self.ti = 0 # The time index, e.g. 0, 1, 2
        self._tvec    = None # The time vector for this instance in date or Dur format
        self._yearvec = None # Time vector as floating point years
        self.initialized = False # Call self.init(sim) to initialise the object
        return

    @property
    def tvec(self):
        if self._tvec is None:
            raise Exception('Timeline object has not yet been not initialized - call `init()` before using the object')
        else:
            return self._tvec

    @property
    def yearvec(self):
        if self._yearvec is None:
            raise Exception('Timeline object has not yet been not initialized - call `init()` before using the object')
        else:
            return self._yearvec

    def __repr__(self):
        if self.initialized:
            return f'<Timeline t={self.tvec[self.ti]}, ti={self.ti}, {self.start}-{self.stop} dt={self.dt}>'
        else:
            return '<Timeline (uninitialized)>'

    def disp(self):
        return sc.pr(self)

    @property
    def timevec(self):
        # Backwards-compatibility function for now - TODO: consider deprecating?
        return self.tvec

    @property
    def datevec(self):
        # Backwards-compatibility function for now - TODO: consider deprecating?
        return self.tvec

    @property
    def npts(self):
        try:
            return self.tvec.shape[0]
        except:
            return 0

    @property
    def dt_year(self):
        return self.dt.years

    def __bool__(self):
        """ Always truthy """
        return True

    def __len__(self):
        """ Length is the number of timepoints """
        return self.npts

    @property
    def ready(self):
        """ Check if all parameters are in place to be initialized """
        return not any([getattr(self, k) is None for k in self.time_args])

    @property
    def finished(self):
        """
        Check if the simulation is finished, i.e. we're at the last time point
        (note, this does not distinguish whether we are at the beginning or end
        of the last time point, so use with caution!)
        """
        return self.ti == self.npts-1

    @property
    def is_absolute(self):
        """
        Check whether the fundamental simulation unit is absolute

        A time vector is absolute if the start is a date rather than a Dur
        A relative time vector can be made absolute by adding a date to it.
        """
        try:
            return isinstance(self.start, date)
        except:
            return False

    def update(self, pars=None, parent=None, reset=True, force=None, **kwargs):
        """ Reconcile different ways of supplying inputs """
        pars = sc.mergedicts(pars)
        stale = False

        for key in self.time_args:
            current_val = getattr(self, key, None)
            parent_val = getattr(parent, key, None)
            kw_val = kwargs.get(key)
            par_val = pars.get(key)

            # Special handling for dt: don't inherit dt if the units are different
            if key == 'dt':
                if isinstance(parent, Timeline):
                    if parent.unit != self.unit:
                        parent_val = 1.0

            if force is False: # Only update missing (None) values
                val = sc.ifelse(current_val, kw_val, par_val, parent_val)
            elif force is None: # Prioritize current value
                val = sc.ifelse(kw_val, par_val, current_val, parent_val)
            elif force is True: # Prioritize parent value
                val = sc.ifelse(kw_val, par_val, parent_val, current_val)
            else:
                errormsg = f'Invalid value {force} for force: must be False, None, or True'
                raise ValueError(errormsg)

            if val != current_val:
                setattr(self, key, val)
                stale = True

        if stale and reset and self.initialized:
            self.init()
        return

    def init(self, sim=None):
        """ Initialize all vectors """

        if sim is not None:
            self.dt    = sc.ifelse(self.dt,    sim.t.dt,    sim.pars.dt)
            self.start = sc.ifelse(self.start, sim.t.start, sim.pars.start)
            self.stop  = sc.ifelse(self.stop,  sim.t.stop,  sim.pars.stop)
            self.dur   = sc.ifelse(self.dur,   sim.t.dur,   sim.pars.dur)

        if sc.isnumber(self.dur):
            self.dur = ss.Dur(self.dur)
        assert self.dur is None or isinstance(self.dur, ss.Dur), 'Timeline.dur must be a number, a Dur object or None'

        def is_calendar_year(val):
            """
            Whether a number should be interpreted as a calendar year -- by default, a number greater than 1900
            """
            return sc.isnumber(val) and val > self.calendar_year_threshold

        match (self.start, self.stop, self.dur):
            case (None, None, None):
                start = self.default_start
                dur = self.default_dur
                stop = start + dur

            case (start, None, None):
                if isinstance(start, ss.Dur):
                    pass  # Already a Dur which is fine
                elif is_calendar_year(start):
                    start = ss.date(start)
                else:
                    start = ss.Dur(start)
                dur = self.default_dur
                stop = start + dur

            case (None, stop, None):
                if isinstance(stop, ss.Dur):
                    start = stop.__class__(0)
                elif is_calendar_year(stop):
                    stop = ss.date(stop)
                    start = ss.date(self.default_start)
                else:
                    stop = ss.Dur(stop)
                    start = ss.Dur(0)
                dur = stop - start

            case (None, None, dur):
                start = self.default_start
                stop = start+dur

            case (start, None, dur):
                if isinstance(start, ss.Dur):
                    pass
                elif is_calendar_year(start):
                    start = ss.Dur(start)
                else:
                    start = ss.date(start)
                stop = start + dur

            case (None, stop, dur):
                if isinstance(stop, ss.Dur):
                    pass
                elif is_calendar_year(stop):
                    stop = ss.date(stop)
                else:
                    stop = ss.Dur(stop)
                start = stop - dur

            case (start, stop, dur):
                # Note that this block will run if dur is None and if it is not None, which is fine because
                # we are ignoring dur in this case (if the user specifies start and stop, they'll be used)
                if sc.isstring(start):
                    start = ss.date(start)
                if sc.isstring(stop):
                    stop = ss.date(stop)

                if sc.isnumber(start) and sc.isnumber(stop):
                    if is_calendar_year(stop):
                        start = ss.date(start)
                        stop = ss.date(stop)
                    else:
                        start = ss.Dur(start)
                        stop = ss.Dur(stop)
                elif sc.isnumber(start):
                    start = stop.__class__(start)
                elif sc.isnumber(stop):
                    stop = start.__class__(stop)
                dur = stop-start
            case _:
                errormsg = f'Failed to match {start = }, {stop = }, and {dur = } to any known pattern. You can use numbers, strings, ss.date, or ss.Dur objects.'
                raise ValueError(errormsg) # This should not occur

        assert isinstance(start, (ss.date, ss.Dur)), 'Start must be date or ss.Dur'
        assert isinstance(stop, (ss.date, ss.Dur)), 'Stop must be a date or Dur'
        assert type(start) is type(stop), 'Start and stop must be the same type'
        assert start <= stop, 'Start must be before stop'

        self.start = start
        self.stop = stop
        self.dur = dur

        if self.dt is None:
            self.dt = self.default_dt

        if sc.isnumber(self.dt):
            self.dt = ss.Dur(self.dt)

        if sim is not None:
            if self.is_absolute != sim.t.is_absolute:
                raise Exception('Cannot mix absolute/relative times across modules')

        if sim is not None and sim.t.initialized and \
                all([type(getattr(self, key)) == type(getattr(sim.t, key)) for key in ('start', 'stop', 'dt')]) and \
                all([getattr(self, key) == getattr(sim.t, key) for key in ('start', 'stop', 'dt')]):
            self._yearvec = sc.dcp(sim.t._yearvec)
            self._tvec    = sc.cp(sim.t._tvec)  # ss.Dates are immutable so only need shallow copy
            self.initialized = True
            return self

        # We need to populate both the tvec (using dates) and the yearvec (using years). However, we
        # need to decide which of these quantities to prioritise considering that the calendar dates
        # don't convert consistently into fractional years due to varying month/year lengths. We will
        # prioritise one or the other depending on what type of quantity the user has specified for dt
        if isinstance(self.dt, ss.years):
            # If dt has been specified as a years then preference setting fractional years. So first
            # calculate the fractional years, and then convert them to the equivalent dates
            self._yearvec = np.round(self.start.years + np.arange(0, self.stop.years - self.start.years + self.dt.years, self.dt.years), 12)  # Subtracting off self.start.years in np.arange increases floating point precision for that part of the operation, reducing the impact of rounding
            if isinstance(self.stop, ss.Dur):
                self._tvec = np.array([self.stop.__class__(x) for x in self._yearvec])
            else:
                self._tvec = np.array([ss.date(x) for x in self._yearvec])
        else:
            # If dt has been specified as a DateDur then preference setting dates. So first
            # calculate the dates/durations, and then convert them to the equivalent fractional years
            if isinstance(self.stop, ss.Dur):
                self._tvec = ss.Dur.arange(self.start, self.stop, self.dt) # TODO: potentially remove/refactor
            else:
                self._tvec = ss.date.arange(self.start, self.stop, self.dt)
            self._yearvec = np.array([x.years for x in self._tvec])

        self.initialized = True
        return self

    def now(self, key=None):
        """
        Get the current simulation time

        Args:
            which (str): which type of time to get: default (None), "year", "date", "tvec", or "str"

        **Examples**:

            t = ss.Timeline(start='2021-01-01', stop='2022-02-02', dt=1, unit='week')
            t.ti = 25
            t.now() # Returns <2021-06-25>
            t.now('date') # Returns <2021-06-25>
            t.now('year') # Returns 2021.479
            t.now('str') # Returns '2021-06-25'
        """
        if key in [None, 'none', 'time', 'str']:
            vec = self.tvec
        elif key == 'year':
            vec = self.yearvec
        else:
            errormsg = f'Invalid key "{key}": must be None, abs, date, or year'
            raise ValueError(errormsg)

        if 0 <= self.ti < len(vec):
            now = vec[self.ti]
        else:
            now = self.tvec[0] + self.dt*self.ti
            if key == 'year':
                now = float(now)

        if key == 'str':
            now = str(now)

        return now



"""
Define the calibration class
"""
import os
import numpy as np
import optuna as op
import pandas as pd
import optuna.visualization.matplotlib as vis
import sciris as sc
import starsim as ss
import matplotlib.pyplot as plt


__all__ = ['Calibration']


class Calibration(sc.prettyobj):
    """
    A class to handle calibration of Starsim simulations. Uses the Optuna hyperparameter
    optimization library (optuna.org).

    Args:
        sim          (Sim)   : the base simulation to calibrate
        calib_pars   (dict)  : a dictionary of the parameters to calibrate of the format ``dict(key1=dict(low=1, high=2, guess=1.5, **kwargs), key2=...)``, where kwargs can include "suggest_type" to choose the suggest method of the trial (e.g. suggest_float) and args passed to the trial suggest function like "log" and "step"
        n_workers    (int)   : the number of parallel workers (if None, will use all available CPUs)
        total_trials (int)   : the total number of trials to run, each worker will run approximately n_trials = total_trial / n_workers
        reseed       (bool)  : whether to generate new random seeds for each trial
        build_fn  (callable) : function that takes a sim object and calib_pars dictionary and returns a modified sim
        build_kw      (dict) : a dictionary of options that are passed to build_fn to aid in modifying the base simulation. The API is ``self.build_fn(sim, calib_pars=calib_pars, **self.build_kw)``, where sim is a copy of the base simulation to be modified with calib_pars
        components    (list) : CalibComponents independently assess pseudo-likelihood as part of evaluating the quality of input parameters
        prune_fn  (callable) : Function that takes a dictionary of parameters and returns True if the trial should be pruned
        eval_fn   (callable) : Function mapping a sim to a float (e.g. negative log likelihood) to be maximized. If None, the default will use CalibComponents.
        eval_kw       (dict) : Additional keyword arguments to pass to the eval_fn
        label        (str)   : a label for this calibration object
        study_name   (str)   : name of the optuna study
        db_name      (str)   : the name of the database file (default: 'starsim_calibration.db')
        continue_db  (bool)  : whether to continue if the database already exists, removes the database if false (default: false, any existing database will be deleted)
        keep_db      (bool)  : whether to keep the database after calibration (default: false, the database will be deleted)
        storage      (str)   : the location of the database (default: sqlite)
        sampler (BaseSampler): the sampler used by optuna, like optuna.samplers.TPESampler
        die          (bool)  : whether to stop if an exception is encountered (default: false)
        debug        (bool)  : if True, do not run in parallel
        verbose      (bool)  : whether to print details of the calibration

    Returns:
        A Calibration object
    """
    def __init__(self, sim, calib_pars, n_workers=None, total_trials=None, reseed=True,
                 build_fn=None, build_kw=None, eval_fn=None, eval_kw=None, components=None, prune_fn=None,
                 label=None, study_name=None, db_name=None, keep_db=None, continue_db=None, storage=None,
                 sampler=None, die=False, debug=False, verbose=True):

        # Handle run arguments
        if total_trials is None: total_trials   = 100
        if n_workers    is None: n_workers      = 1 if debug else sc.cpu_count()
        if study_name   is None: study_name     = 'starsim_calibration'
        if db_name      is None: db_name        = f'{study_name}.db'
        if continue_db  is None: continue_db    = False
        if keep_db      is None: keep_db        = False
        if storage      is None: storage        = f'sqlite:///{db_name}'
        
        self.build_fn       = build_fn
        self.build_kw       = build_kw or dict()
        self.eval_fn        = eval_fn or self._eval_fit
        self.eval_kw        = eval_kw or dict()
        self.components     = sc.tolist(components)
        self.prune_fn       = prune_fn

        n_trials = int(np.ceil(total_trials/n_workers))
        kw = dict(n_trials=n_trials, n_workers=int(n_workers), debug=debug, study_name=study_name,
                  db_name=db_name, continue_db=continue_db, keep_db=keep_db, storage=storage, sampler=sampler)
        self.run_args = sc.objdict(kw)

        # Handle other inputs
        self.label      = label
        self.sim        = sim
        self.calib_pars = calib_pars
        self.reseed     = reseed
        self.die        = die
        self.verbose    = verbose
        self.calibrated = False
        self.before_msim = None
        self.after_msim  = None

        self.study = None

        return

    def run_sim(self, calib_pars=None, label=None):
        """ Create and run a simulation """
        sim = sc.dcp(self.sim)
        if label: sim.label = label

        sim = self.build_fn(sim, calib_pars=calib_pars, **self.build_kw)

        try:
            sim.run() # Run the simulation (or MultiSim)
            return sim
        except Exception as E:
            if self.die:
                raise E
            else:
                print(f'Encountered error running sim!\nParameters:\n{calib_pars}\nTraceback:\n{sc.traceback()}')
                output = None
                return output

    def _sample_from_trial(self, pardict=None, trial=None):
        """
        Take in an optuna trial and sample from pars, after extracting them from
        the structure they're provided in
        """
        pars = sc.dcp(pardict)
        for parname, spec in pars.items():
            if 'value' in spec:
                # Already have a value, likely running initial or final values as part of checking the fit
                continue

            if 'suggest_type' in spec:
                suggest_type = spec.pop('suggest_type')
                sampler_fn = getattr(trial, suggest_type)
            else:
                sampler_fn = trial.suggest_float

            path = spec.pop('path', None) # remove path for the sampler
            guess = spec.pop('guess', None) # remove guess for the sampler
            spec['value'] = sampler_fn(name=parname, **spec) # suggest values!
            spec['path'] = path
            spec['guess'] = guess

        return pars

    def _eval_fit(self, sim, **kwargs):
        """ Evaluate the fit by evaluating the negative log likelihood, used only for components"""
        nll = 0 # Negative log likelihood
        for component in self.components:
            nll += component(sim, **kwargs)
        return nll

    def plot(self, **kwargs):
        """"
        Plot the calibration results. For a component-based likelihood, it only
        makes sense to directly call plot after calling eval_fn.
        """
        assert self.before_msim is not None and self.after_msim is not None, 'Please run check_fit() before plotting'
        figs = []
        for component in self.components:
            component.eval(self.before_msim)
            before_actual = component.actual
            before_actual['calibrated'] = 'Before Calibration'

            component.eval(self.after_msim)
            after_actual = component.actual
            after_actual['calibrated'] = 'After Calibration'

            actual = pd.concat([before_actual, after_actual])
            fig = component.plot(actual, **kwargs)
            figs.append(fig)
        return figs

    def run_trial(self, trial):
        """ Define the objective for Optuna """
        if self.calib_pars is not None:
            pars = self._sample_from_trial(self.calib_pars, trial)
        else:
            pars = None

        if self.reseed:
            pars['rand_seed'] = trial.suggest_int('rand_seed', 0, 1_000_000) # Choose a random rand_seed

        # Prune if the prune_fn returns True
        if self.prune_fn is not None and self.prune_fn(pars):
            raise op.exceptions.TrialPruned()

        sim = self.run_sim(pars)

        # Compute fit
        fit = self.eval_fn(sim, **self.eval_kw)
        return fit

    def worker(self):
        """ Run a single worker """

        if self.verbose:
            op.logging.set_verbosity(op.logging.DEBUG)
        else:
            op.logging.set_verbosity(op.logging.ERROR)
        study = op.load_study(storage=self.run_args.storage, study_name=self.run_args.study_name, sampler=self.run_args.sampler)
        output = study.optimize(self.run_trial, n_trials=self.run_args.n_trials, callbacks=None)
        return output

    def run_workers(self):
        """ Run multiple workers in parallel """
        if self.run_args.n_workers > 1 and not self.run_args.debug: # Normal use case: run in parallel
            output = sc.parallelize(self.worker, iterarg=self.run_args.n_workers)
        else: # Special case: just run one
            output = [self.worker()]
        return output

    def remove_db(self):
        """ Remove the database file if keep_db is false and the path exists """
        try:
            if 'sqlite' in self.run_args.storage:
                # Delete the file from disk
                if os.path.exists(self.run_args.db_name):
                    os.remove(self.run_args.db_name)
                if self.verbose: print(f'Removed existing calibration file {self.run_args.db_name}')
            else:
                # Delete the study from the database e.g., mysql
                op.delete_study(study_name=self.run_args.study_name, storage=self.run_args.storage)
                if self.verbose: print(f'Deleted study {self.run_args.study_name} in {self.run_args.storage}')
        except Exception as E:
            if self.verbose:
                print('Could not delete study, skipping...')
                print(str(E))
        return

    def make_study(self):
        """ Make a study, deleting if it already exists and user does not want to continue_db """
        if not self.run_args.continue_db:
            self.remove_db()
        if self.verbose: print(self.run_args.storage)
        try:
            study = op.create_study(storage=self.run_args.storage, study_name=self.run_args.study_name, direction='minimize')
        except op.exceptions.DuplicatedStudyError:
            ss.warn(f'Study named {self.run_args.study_name} already exists in storage {self.run_args.storage}, loading...')
            study = op.create_study(storage=self.run_args.storage, study_name=self.run_args.study_name, direction='minimize', load_if_exists=True)
            try:
                self.best_pars = sc.objdict(study.best_params)
            except Exception as E:
                print(f'Could not get best parameters: {str(E)}')
                self.best_pars = None
        return study

    def calibrate(self, calib_pars=None, **kwargs):
        """
        Perform calibration.

        Args:
            calib_pars (dict): if supplied, overwrite stored calib_pars
            kwargs (dict): if supplied, overwrite stored run_args (n_trials, n_workers, etc.)
        """
        # Load and validate calibration parameters
        if calib_pars is not None:
            self.calib_pars = calib_pars
        self.run_args.update(kwargs) # Update optuna settings

        # Run the optimization
        t0 = sc.tic()
        self.study = self.make_study()
        self.run_workers()
        study = op.load_study(storage=self.run_args.storage, study_name=self.run_args.study_name, sampler=self.run_args.sampler)
        self.best_pars = sc.objdict(study.best_params)
        self.elapsed = sc.toc(t0, output=True)

        # Parse the study into a data frame, self.df while also storing the best parameters
        self.parse_study(study)

        if self.verbose: print('Best pars:', self.best_pars)

        # Tidy up
        self.calibrated = True
        if not self.run_args.keep_db:
            self.remove_db()

        return self

    def to_df(self, top_k=None):
        """ Return the top K results as a dataframe, sorted by value """
        if self.study is None:
            raise ValueError('Please run calibrate() before saving results')

        df = sc.dataframe(self.study.trials_dataframe())
        df = df.sort_values(by='value').set_index('number')

        if top_k is not None:
            df = df.head(top_k)

        return df

    def check_fit(self, do_plot=True):
        """ Run before and after simulations to validate the fit """
        if self.verbose: sc.printcyan('\nChecking fit...')

        before_pars = sc.dcp(self.calib_pars)
        for spec in before_pars.values():
            spec['value'] = spec['guess'] # Use guess values

        # Load in case calibration was interrupted
        if self.best_pars is None:
            try:
                study = op.load_study(storage=self.run_args.storage, study_name=self.run_args.study_name, sampler=self.run_args.sampler)
                self.best_pars = sc.objdict(study.best_params)
            except:
                raise ValueError('Seems like calibration did not finish successfully and also unable to obtain best parameters from the {self.run_args.storage}:{self.run_args.study_name} as the study was likely automatically deleted, see keep_db.')

        after_pars = sc.dcp(self.calib_pars)
        for parname, spec in after_pars.items():
            spec['value'] = self.best_pars[parname] # Use best parameters from calibration

        self.before_msim = self.build_fn(self.sim.copy(), calib_pars=before_pars, **self.build_kw)
        self.after_msim = self.build_fn(self.sim.copy(), calib_pars=after_pars, **self.build_kw)

        fix_before = isinstance(self.before_msim, ss.Sim)
        fix_after = isinstance(self.after_msim, ss.Sim)
        if fix_after or fix_after:
            if fix_before:
                self.before_msim = ss.MultiSim(self.before_msim, initialize=True, debug=True, parallel=False, n_runs=1)

            if fix_after:
                self.after_msim = ss.MultiSim(self.after_msim, initialize=True, debug=True, parallel=False, n_runs=1)

        msim = ss.MultiSim(self.before_msim.sims + self.after_msim.sims)
        msim.run()

        self.before_fits = self.eval_fn(self.before_msim, **self.eval_kw)
        self.after_fits = self.eval_fn(self.after_msim, **self.eval_kw)

        if do_plot:
            figs = self.plot()

        print(f'Fit with original pars: {self.before_fits}')
        print(f'Fit with best-fit pars: {self.after_fits}')
        
        before = self.before_fits.mean() if isinstance(self.before_fits, np.ndarray) else self.before_fits
        after = self.after_fits.mean() if isinstance(self.after_fits, np.ndarray) else self.after_fits

        if after <= before:
            print(f'✓ Calibration improved fit {before} --> {after}')
            return True
        
        print(f'✗ Calibration did not improve fit as the objective got worse ({before} --> {after}), but this sometimes happens stochastically and is not necessarily an error')
        return False

    def parse_study(self, study):
        """Parse the study into a data frame -- called automatically """
        best = study.best_params
        self.best_pars = best

        if self.verbose: print('Making results structure...')
        results = []
        n_trials = len(study.trials)
        failed_trials = []
        for trial in study.trials:
            data = {'index':trial.number, 'mismatch': trial.value}
            for key,val in trial.params.items():
                data[key] = val
            if data['mismatch'] is None:
                failed_trials.append(data['index'])
            else:
                results.append(data)
        if self.verbose: print(f'Processed {n_trials} trials; {len(failed_trials)} failed')

        keys = ['index', 'mismatch'] + list(best.keys())
        data = sc.objdict().make(keys=keys, vals=[])
        for i,r in enumerate(results):
            for key in keys:
                if key not in r:
                    warnmsg = f'Key {key} is missing from trial {i}, replacing with default'
                    print(warnmsg)
                    r[key] = best[key]
                data[key].append(r[key])
        self.study_data = data
        self.df = sc.dataframe.from_dict(data)
        self.df = self.df.sort_values(by=['mismatch']) # Sort
        return

    def to_json(self, filename=None, indent=2, **kwargs):
        """ Convert the results to JSON """
        order = np.argsort(self.df['mismatch'])
        json = []
        for o in order:
            row = self.df.iloc[o,:].to_dict()
            rowdict = dict(index=row.pop('index'), mismatch=row.pop('mismatch'), pars={})
            for key,val in row.items():
                rowdict['pars'][key] = val
            json.append(rowdict)
        self.json = json
        if filename:
            return sc.savejson(filename, json, indent=indent, **kwargs)
        else:
            return json

    def plot_final(self, **kwargs):
        """
        Plot sims after calibration

        Args:
            kwargs (dict): passed to MultiSim.plot()
        """

        jup = ss.options.jupyter if 'jupyter' in ss.options else sc.isjupyter()
        ss.options.jupyter = False

        pars = sc.dcp(self.calib_pars)
        for parname, spec in pars.items():
            spec['value'] = self.best_pars[parname] # Use best parameters from calibration
        msim = self.build_fn(self.sim.copy(), calib_pars=pars, **self.build_kw)

        msim.run()
        fits = self.eval_fn(msim, **self.eval_kw)

        if isinstance(msim, ss.MultiSim): # It could be a single simulation
            msim.reduce()
        fig = msim.plot()
        fig.suptitle('After calibration')

        ss.options.jupyter = jup
        return fig

    def plot_optuna(self, methods=None):
        """ Plot Optuna's visualizations """
        figs = []

        methods = sc.tolist(methods)

        if methods is None:
            methods = [
                'plot_contour',
                'plot_edf',
                'plot_hypervolume_history',
                'plot_intermediate_values',
                'plot_optimization_history',
                'plot_parallel_coordinate',
                'plot_param_importances',
                'plot_pareto_front',
                'plot_rank',
                'plot_slice',
                'plot_terminator_improvement',
                'plot_timeline',
            ]

        for method in methods:
            try:
                fig = getattr(vis, method)(self.study)
                figs.append(fig)
            except Exception as E:
                print(f'Could not run {method}: {str(E)}')
        return figs



"""
Exploring singlerng vs rng for a model HIV system sweeping coverage of ART.
"""

# %% Imports and settings
import os
import starsim as ss
import scipy.stats as sps
import sciris as sc
import pandas as pd
import seaborn as sns
import numpy as np

# Suppress warning from seaborn
import warnings
warnings.filterwarnings("ignore", "is_categorical_dtype")
warnings.filterwarnings("ignore", "use_inf_as_na")

rngs = ['centralized', 'single', 'multi']

n = 100 # Agents
n_rand_seeds = 25
intv_cov_levels = [0.01, 0.10, 0.25, 0.73] + [0] # Must include 0 as that's the baseline

figdir = os.path.join(os.getcwd(), 'figs', 'ART')
sc.path(figdir).mkdir(parents=True, exist_ok=True)

def run_sim(n, idx, intv_cov, rand_seed, rng):

    print(f'Starting sim {idx} with rand_seed={rand_seed} and intv_cov={intv_cov}, rng={rng}')

    ppl = ss.People(n)

    en_pars = dict(duration=sps.weibull_min(c=1.5, scale=3)) # c is shape, gives a mean of about 0.9*scale years.
    networks = ss.ndict(ss.EmbeddingNet(en_pars), ss.MaternalNet())

    hiv_pars = {
        'beta': {'embedding': [0.08, 0.06], 'maternal': [0.3, 0]},
        'init_prev': np.maximum(5/n, 0.01),
        'art_efficacy': 0.96,
        'death_prob': 0.06,
    }
    hiv = ss.HIV(hiv_pars)

    pregnancy = ss.Pregnancy(pars={'fertility_rate': 10})
    deaths = ss.Deaths()

    pars = {
        'start': 1980,
        'end': 2070,
        'dt': 0.1,
        'rand_seed': rand_seed,
        'verbose': 0,
        'remove_dead': True,
        'slot_scale': 10 # Increase slot scale to reduce repeated slots
    }

    if intv_cov > 0:
        pars['interventions'] = [ ss.hiv.ART(year=[2004, 2020], coverage=[0, intv_cov]) ]

    sim = ss.Sim(people=ppl, networks=networks, diseases=[hiv], demographics=[pregnancy, deaths], pars=pars,
            label=f'Sim with {n} agents and intv_cov={intv_cov}')
    sim.initialize()
    sim.run()

    df = pd.DataFrame( {
        'year': sim.yearvec,
        #'hiv.n_infected': sim.results.hiv.n_infected, # Optional, but mostly redundant with prevalence
        'hiv.prevalence': sim.results.hiv.prevalence,
        'hiv.cum_deaths': sim.results.hiv.new_deaths.cumsum(),
        'pregnancy.cum_births': sim.results.pregnancy.births.cumsum(),
    })
    df['intv_cov'] = intv_cov
    df['rand_seed'] = rand_seed
    df['rng'] = rng

    print(f'Finishing sim {idx} with rand_seed={rand_seed} and intv_cov={intv_cov}, rng={rng}')

    return df

def run_scenarios():
    results = []
    times = {}
    for rng in rngs:
        ss.options(rng=rng)
        cfgs = []
        for rs in range(n_rand_seeds):
            for intv_cov in intv_cov_levels:
                cfgs.append({'intv_cov':intv_cov, 'rand_seed':rs, 'rng':rng, 'idx':len(cfgs)})
        T = sc.tic()
        results += sc.parallelize(run_sim, kwargs={'n': n}, iterkwargs=cfgs, die=True, serial=False)
        times[f'rng={rng}'] = sc.toc(T, output=True)

    print('Timings:', times)

    df = pd.concat(results)
    df.to_csv(os.path.join(figdir, 'result.csv'))
    return df

def plot_scenarios(df):
    d = pd.melt(df, id_vars=['year', 'rand_seed', 'intv_cov', 'rng'], var_name='channel', value_name='Value')
    d['baseline'] = d['intv_cov']==0
    bl = d.loc[d['baseline']]
    scn = d.loc[~d['baseline']]
    bl = bl.set_index(['year', 'channel', 'rand_seed', 'intv_cov', 'rng'])[['Value']].reset_index('intv_cov')
    scn = scn.set_index(['year', 'channel', 'rand_seed', 'intv_cov', 'rng'])[['Value']].reset_index('intv_cov')
    mrg = scn.merge(bl, on=['year', 'channel', 'rand_seed', 'rng'], suffixes=('', '_ref'))
    mrg['Value - Reference'] = mrg['Value'] - mrg['Value_ref']
    mrg = mrg.sort_index()

    fkw = {'sharey': False, 'sharex': 'col', 'margin_titles': True}

    ## TIMESERIES
    g = sns.relplot(kind='line', data=d, x='year', y='Value', hue='intv_cov', col='channel', row='rng', row_order=rngs,
        height=5, aspect=1.2, palette='Set1', errorbar='sd', lw=2, facet_kws=fkw)
    g.set_titles(col_template='{col_name}', row_template='rng: {row_name}')
    g.set_xlabels(r'$t_i$')
    g.figure.savefig(os.path.join(figdir, 'timeseries.png'), bbox_inches='tight', dpi=300)

    ## DIFF TIMESERIES
    for ms, mrg_by_ms in mrg.groupby('rng'):
        g = sns.relplot(kind='line', data=mrg_by_ms, x='year', y='Value - Reference', hue='intv_cov', col='channel',
                row='intv_cov', height=3, aspect=1.0, palette='Set1', estimator=None, units='rand_seed', lw=0.5, facet_kws=fkw) #errorbar='sd', lw=2, 
        g.set_titles(col_template='{col_name}', row_template='Coverage: {row_name}')
        #g.figure.suptitle('MultiRNG' if ms else 'SingleRNG')
        g.figure.subplots_adjust(top=0.88)
        g.set_xlabels('Year')
        g.figure.savefig(os.path.join(figdir, f'diff_{ms}.png'), bbox_inches='tight', dpi=300)

    ## FINAL TIME
    tf = df['year'].max()
    mtf = mrg.loc[tf]
    g = sns.displot(data=mtf.reset_index(), kind='kde', fill=True, rug=True, cut=0, hue='intv_cov', x='Value - Reference',
            col='channel', row='rng', row_order=rngs, height=5, aspect=1.2, facet_kws=fkw, palette='Set1')
    g.set_titles(col_template='{col_name}', row_template='rng: {row_name}')
    g.set_xlabels(f'Value - Reference at year {tf}')
    g.fig.savefig(os.path.join(figdir, 'final.png'), bbox_inches='tight', dpi=300)

    print('Figures saved to:', os.path.join(os.getcwd(), figdir))

    return


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--plot', help='plot from a cached CSV file', type=str)
    args = parser.parse_args()

    if args.plot:
        print('Reading CSV file', args.plot)
        df = pd.read_csv(args.plot, index_col=0)
    else:
        print('Running scenarios')
        df = run_scenarios()

    plot_scenarios(df)
    print('Done')



"""
Base classes for diseases
"""
import numpy as np
import sciris as sc
import starsim as ss
import networkx as nx
from operator import itemgetter
import pandas as pd

ss_int_ = ss.dtypes.int
ss_float_ = ss.dtypes.float

__all__ = ['Disease', 'Infection', 'InfectionLog']


class Disease(ss.Module):
    """ Base module class for diseases """

    def __init__(self, pars=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.define_pars(
            log = False,
        )
        self.update_pars(pars, **kwargs)
        self.results = ss.Results(self.name)
        return

    @property
    def _disease_states(self):
        """
        Iterator over disease states with boolean type

        For diseases, these states typically represent attributes like 'susceptible',
        'infectious', 'diagnosed' etc. These variables are typically useful to store
        results for.
        """
        for state in self.states:
            if isinstance(state, ss.State):
                yield state
        return

    def init_pre(self, sim):
        """ Link the disease to the sim, create objects, and initialize results; see Module.init_pre() for details """
        super().init_pre(sim)
        if self.pars.log:
            self.log = InfectionLog()
        return

    def init_results(self):
        """
        Initialize results

        By default, diseases all report on counts for any explicitly defined "States", e.g. if
        a disease contains a boolean state 'susceptible' it will automatically contain a
        Result for 'n_susceptible'.
        """
        super().init_results()
        results = sc.autolist()
        for state in self._disease_states:
            results += ss.Result(f'n_{state.name}', dtype=int, scale=True, label=state.label)
        self.define_results(*results)
        return

    def step_state(self):
        """
        Carry out updates at the start of the timestep (prior to transmission);
        these are typically state changes
        """
        pass

    def step_die(self, uids):
        """
        Carry out state changes upon death

        This function is triggered after deaths are resolved, and before analyzers are run.
        See the SIR example model for a typical use case - deaths are requested as an autonomous
        update, to take effect after transmission on the same timestep. State changes that occur
        upon death (e.g., clearing an `infected` flag) are executed in this function. That also
        allows an intervention to avert a death scheduled on the same timestep, without having
        to undo any state changes that have already been applied (because they only run via this
        function if the death actually occurs).

        Unlike other methods during the integration loop, this method is not called directly
        by the sim; instead, it is called by people.step_die(), which reconciles the UIDs of
        the agents who will die.

        Depending on the module and the results it produces, it may or may not be necessary
        to implement this.
        """
        pass

    def step(self):
        """
        Handle the main disease updates, e.g. add new cases

        This method is agnostic as to the mechanism by which new cases occur. This
        could be through transmission (parametrized in different ways, which may or
        may not use the contact networks) or it may be based on risk factors/seeding,
        as may be the case for non-communicable diseases.

        It is expected that this method will internally call Disease.set_prognoses()
        at some point.
        """
        pass

    def set_prognoses(self, uids, sources=None):
        """
        Set prognoses upon infection/acquisition

        This function assigns state values upon infection or acquisition of
        the disease. It would normally be called somewhere towards the end of
        `Disease.make_new_cases()`. Infections will optionally be added to
        the log as part of this operation if logging is enabled (in the
        `Disease` parameters)

        The `sources` are relevant for infectious diseases, but would be left
        as `None` for NCDs.

        Args:
            sim (Sim): the STarsim simulation object
            uids (array): UIDs for agents to assign disease progoses to
            from_uids (array): Optionally specify the infecting agent
        """
        # Track infections
        if self.pars.log:
            self.log_infections(uids, sources)
        return

    def log_infections(self, uids, sources=None):
        self.log.add_entries(uids, sources, self.now)
        return

    def update_results(self):
        """
        Update results

        This function is executed after transmission in all modules has been resolved.
        This allows result updates at this point to capture outcomes dependent on multiple
        modules, where relevant.
        """
        for state in self._disease_states:
            self.results[f'n_{state.name}'][self.ti] = state.sum()
        return


class Infection(Disease):
    """
    Base class for infectious diseases used in Starsim

    This class contains specializations for infectious transmission (i.e., implements network-based
    transmission with directional beta values) and defines attributes that connectors
    operate on to capture co-infection
    """

    def __init__(self, pars=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.define_states(
            ss.State('susceptible', default=True, label='Susceptible'),
            ss.State('infected', label='Infected'),
            ss.FloatArr('rel_sus', default=1.0, label='Relative susceptibility'),
            ss.FloatArr('rel_trans', default=1.0, label='Relative transmission'),
            ss.FloatArr('ti_infected', label='Time of infection' ),
        )

        self.define_pars(
            init_prev = None, # Replace None with a ss.bernoulli to seed infections
        )
        self.update_pars(pars, **kwargs)

        # Define random number generator for determining transmission
        self.trans_rng = ss.multi_random('source', 'target')
        return

    def init_pre(self, sim):
        super().init_pre(sim)
        self.validate_beta(run_checks=True)
        return

    @property
    def infectious(self):
        """
        Generally defined as an alias for infected, although these may differ in some diseases.
        Transmission comes from infectious people; prevalence estimates may include infected people who don't transmit
        """
        return self.infected

    def init_post(self):
        """
        Set initial values for states. This could involve passing in a full set of initial conditions,
        or using init_prev, or other. Note that this is different to initialization of the Arr objects
        i.e., creating their dynamic array, linking them to a People instance. That should have already
        taken place by the time this method is called.
        """
        super().init_post()
        if self.pars.init_prev is None:
            return

        initial_cases = self.pars.init_prev.filter()
        if len(initial_cases):
            self.set_prognoses(initial_cases)  # TODO: sentinel value to indicate seeds?

        # Exclude initial cases from results -- disabling for now since it disrupts counting of new infections, e.g. test_diseases.py
        # self.ti_infected[self.ti_infected == self.ti] = -1
        return initial_cases

    def init_results(self):
        """
        Initialize results
        """
        super().init_results()
        self.define_results(
            ss.Result('prevalence',     dtype=float, scale=False, label='Prevalence'),
            ss.Result('new_infections', dtype=int,   scale=True,  label='New infections'),
            ss.Result('cum_infections', dtype=int,   scale=True,  label='Cumulative infections'),
        )
        return

    def validate_beta(self, run_checks=False):
        """ Validate beta and return as a map to match the networks """
        sim = self.sim
        β = self.pars.beta

        def scalar_beta(β):
            return isinstance(β, ss.TimePar) or sc.isnumber(β)

        if run_checks:
            scalar_warn = f'Beta is defined as a number ({β}); convert it to a rate to handle timestep conversions'

            if 'beta' not in self.pars:
                errormsg = f'Disease {self.name} is missing beta; pars are: {sc.strjoin(self.pars.keys())}'
                raise sc.KeyNotFoundError(errormsg)

            if sc.isnumber(β):
                ss.warn(scalar_warn)
            elif isinstance(β, dict):
                for netbeta in β.values():
                    if sc.isnumber(netbeta):
                        ss.warn(scalar_warn)
                    elif isinstance(netbeta, (list, tuple)):
                        for thisbeta in netbeta:
                            if sc.isnumber(netbeta):
                                ss.warn(scalar_warn)

        # If beta is a scalar, apply this bi-directionally to all networks
        if scalar_beta(β):
            betamap = {ss.standardize_netkey(k):[β,β] for k in sim.networks.keys()}

        # If beta is a dict, check all entries are bi-directional
        elif isinstance(β, dict):
            betamap = dict()
            for k,thisbeta in β.items():
                nkey = ss.standardize_netkey(k)
                if scalar_beta(thisbeta):
                    betamap[nkey] = [thisbeta, thisbeta]
                else:
                    betamap[nkey] = thisbeta

        else:
            errormsg = f'Invalid type {type(β)} for beta'
            raise TypeError(errormsg)

        # Check that it matches the network
        netkeys = [ss.standardize_netkey(k) for k in list(sim.networks.keys())]
        if set(betamap.keys()) != set(netkeys):
            errormsg = f'Network keys ({netkeys}) and beta keys ({betamap.keys()}) do not match'
            raise ValueError(errormsg)

        return betamap

    def step(self):
        """
        Perform key infection updates, including infection and setting prognoses
        """
        # Create new cases
        new_cases, sources, networks = self.infect() # TODO: store outputs in self or use objdict rather than 3 returns

        # Set prognoses
        if len(new_cases):
            self.set_outcomes(new_cases, sources)

        return new_cases, sources, networks

    @staticmethod # In future, consider: @nb.njit(fastmath=True, parallel=True, cache=True), but no faster it seems
    def compute_transmission(src, trg, rel_trans, rel_sus, beta_per_dt, randvals):
        """ Compute the probability of a->b transmission """
        p_transmit = rel_trans[src] * rel_sus[trg] * beta_per_dt
        transmitted = p_transmit > randvals
        target_uids = trg[transmitted]
        source_uids = src[transmitted]
        return target_uids, source_uids

    def infect(self):
        """ Determine who gets infected on this timestep via transmission on the network """
        new_cases = []
        sources = []
        networks = []
        betamap = self.validate_beta()

        rel_trans = self.rel_trans.asnew(self.infectious * self.rel_trans)
        rel_sus   = self.rel_sus.asnew(self.susceptible * self.rel_sus)

        for i, (nkey,route) in enumerate(self.sim.networks.items()):
            nk = ss.standardize_netkey(nkey)
            if isinstance(route, (ss.MixingPool, ss.MixingPools)):
                target_uids = route.compute_transmission(rel_sus, rel_trans, betamap[nk])
                new_cases.append(target_uids)
                sources.append(np.full(len(target_uids), dtype=ss_float_, fill_value=np.nan))
                networks.append(np.full(len(target_uids), dtype=ss_int_, fill_value=i))
            elif isinstance(route, ss.Network) and len(route): # Skip networks with no edges
                edges = route.edges
                p1p2b0 = [edges.p1, edges.p2, betamap[nk][0]] # Person 1, person 2, beta 0
                p2p1b1 = [edges.p2, edges.p1, betamap[nk][1]] # Person 2, person 1, beta 1
                for src, trg, beta in [p1p2b0, p2p1b1]:
                    if beta: # Skip networks with no transmission
                        beta_per_dt = route.net_beta(disease_beta=beta, disease=self) # Compute beta for this network and timestep
                        randvals = self.trans_rng.rvs(src, trg) # Generate a new random number based on the two other random numbers
                        args = (src, trg, rel_trans, rel_sus, beta_per_dt, randvals) # Set up the arguments to calculate transmission
                        target_uids, source_uids = self.compute_transmission(*args) # Actually calculate it
                        new_cases.append(target_uids)
                        sources.append(source_uids)
                        networks.append(np.full(len(target_uids), dtype=ss_int_, fill_value=i))

        # Finalize
        if len(new_cases) and len(sources):
            new_cases = ss.uids.cat(new_cases)
            new_cases, inds = new_cases.unique(return_index=True)
            sources = ss.uids.cat(sources)[inds]
            networks = np.concatenate(networks)[inds]
        else:
            new_cases = ss.uids()
            sources = ss.uids()
            networks = np.empty(0, dtype=ss_int_)

        return new_cases, sources, networks

    def set_outcomes(self, uids, sources=None):
        sim = self.sim
        congenital = sim.people.age[uids] <= 0
        if np.count_nonzero(congenital):
            src_c = sources[congenital] if sources is not None else None
            self.set_congenital(uids[congenital], src_c)
        src_p = sources[~congenital] if sources is not None else None
        self.set_prognoses(uids[~congenital], src_p)
        return

    def set_congenital(self, uids, sources=None):
        pass

    def update_results(self):
        super().update_results()
        res = self.results
        ti = self.ti
        res.prevalence[ti] = res.n_infected[ti] / np.count_nonzero(self.sim.people.alive)
        res.new_infections[ti] = np.count_nonzero(np.round(self.ti_infected) == ti)
        res.cum_infections[ti] = np.sum(res['new_infections'][:ti+1]) # TODO: can compute at end
        return


class InfectionLog(nx.MultiDiGraph):
    """
    Record infections

    The infection log records transmission events and optionally other data
    associated with each transmission. Basic functionality is to track
    transmission with

    >>> Disease.log.append(source, target, t)

    Seed infections can be recorded with a source of `None`, although all infections
    should have a target and a time. Other data can be captured in the log, either at
    the time of creation, or later on. For example

    >>> Disease.log.append(source, target, t, network='msm')

    could be used by a module to track the network in which transmission took place.
    Modules can optionally add per-infection outcomes later as well, for example

    >>> Disease.log.add_data(source, t_dead=2024.25)

    This would be equivalent to having specified the data at the original time the log
    entry was created - however, it is more useful for tracking events that may or may
    not occur after the infection and could be modified by interventions (e.g., tracking
    diagnosis, treatment, notification etc.)

    A table of outcomes can be returned using `InfectionLog.line_list()`
    """
    def add_entries(self, uids, sources=None, time=np.nan):
        if sources is None:
            for target in uids:
                self.append(np.nan, target, time)
        else:
            for target, source in zip(uids, sources):
                self.append(source, target, time)
        return

    def add_data(self, uids, **kwargs):
        """
        Record extra infection data

        This method can be used to add data to an existing transmission event.
        The most recent transmission event will be used

        Args:
            uid: The UID of the target node (the agent that was infected)
            kwargs: Remaining arguments are stored as edge data
        """
        for uid in sc.toarray(uids):
            source, target, key = max(self.in_edges(uid, keys=True), key=itemgetter(2, 0))  # itemgetter twice as fast as lambda apparently
            self[source][target][key].update(**kwargs)
        return

    def append(self, source, target, t, **kwargs):
        self.add_edge(source, target, key=t, **kwargs)
        return

    def line_list(self):
        """
        Return a tabular representation of the log

        This function returns a dataframe containing columns for all quantities
        recorded in the log. Note that the log will contain `NaN` for quantities
        that are defined for some edges and not others (and which are missing for
        a particular entry)
        """
        if len(self) == 0:
            return sc.dataframe(columns=['t', 'source', 'target'])

        entries = []
        for source, target, t, data in self.edges(keys=True, data=True):
            d = data.copy()
            d.update(source=source, target=target, t=t)
            entries.append(d)
        df = sc.dataframe.from_records(entries)
        df = df.sort_values(['t', 'source', 'target'])
        df = df.reset_index(drop=True)

        # Use Pandas "Int64" type to allow nullable integers. This allows the 'source' column
        # to have an integer type corresponding to UIDs while simultaneously supporting the use
        # of null values to represent exogenous/seed infections
        df = df.fillna(pd.NA)
        df['source'] = df['source'].astype("Int64")
        df['target'] = df['target'].astype("Int64")

        return df



"""
Vignette 10: Custom people
"""

import stisim as ss
import numpy as np

##################
# @RomeshA
##################
#Custom people attributes

# Make a new module

# Derive from people
class MyPeople(ss.People):
    def __init__(self):
        super().__init__()
        self.states['risk_tolerance'] = State('risk_tolerance', float, False)
        self.something_else = 'bar'

    def update_pre(self):
        super().update_pre()
        <do something related to the new attributes>
        <or could overload another function>


##################
# @pausz
##################
# The People class in STIsim is a crucial component that handles various operations related to the simulated population.
# In this vignette, we show how to extend and customize the People class by adding new attributes and methods.

# Case 0: Add custom method to people
class CustomPeople(ss.People):
    """
    A custom class that extends the People class to add new functionality.

    Args:
        states (dict, optional): A dictionary of custom states to add to the people.

    **Examples**::
        ppl = CustomPeople(states=new_states)
    """

    def __init__(self, states=None, *args, **kwargs):
        """ Initialize the CustomPeople object. """
        super().__init__(*args, states=states, **kwargs)

    def plot_state(self, states, **kwargs):
        """
        Plot the distribution of specific states among the population.
        This method could plot all pairs of states as 2D scatterplots, or 2D histograms,
        to detect associations. If the state is boolean, then plot a swarmplot.
        """
        # Handle grid of subplots if multiple are passed
        self.plot_single_state()


# Case 1: Add new states to people
new_states = [ss.State('is_vaccinated', bool, False),                      # Everyone starts unvaccinated
              ss.State('ti_vaccinated', ss.default_float, np.nan),         # Track when a person is vaccinated
              ss.StochState('fertility_level', ss.default_float,           # Fertility levels
                            distdict=dict(dist='normal',
                                          par1=[0.5],
                                          par2=[0.1])),
              ss.State('infertile', bool, False),                          # Add a new base state that tracks who is fertile
              ss.State('ti_infertile', ss.default_float, np.nan),
              ss.StochState('has_access_to_medical_care', bool,            # Add a new base state
                            distdict=dict(dist='choice',
                                          par1=[1, 0],
                                          par2=[0.45, 0.55]))
]



# Create the people with new states
ppl = ss.CustomPeople(states=new_states)

# Once we have customised the People class by adding new base states,
# we can utilise them in the simulation.
sim = ss.Sim(people=ppl,
             network_structure='random',
             modules=ss.gonorrhea())

# Accessing and utilising custom attributes
sim.ppl.plot_state(['fertilty_rate', 'has_access_to_medical_care')


##################
# @robynstuart
##################
# See example in v09_custom_networks.py, i.e.
is_fsw = ss.StochState('fsw', bool, distdict=dict(dist='choice', par1=[1, 0], par2=[0.05, 0.95]))
is_client = ss.StochState('client', bool, distdict=dict(dist='choice', par1=[1, 0], par2=[0.15, 0.85]))
ppl = ss.People(100, states=[is_fsw, is_client])
sex_work = sex_work()  # Question: what constraints are there on this? Must it be a ss.Network or could it be anything?
syph = ss.syphilis()
sim = ss.Sim(people=ppl, networks=['mf', sex_work], modules=syph)
